import yaml
import re
import os


def _parse_docstring_parameters(doc_string, search_param_section=True):
    r"""
    Parses Google-style docstring and returns the list of (``parameter_name``, ``parameter_description``)
    pairs. The properly composed parameter section of the docstring should start with the line
    containing the word ``    Parameters`` followed by the line containing ``    ----------`` and
    end with the line containing the word ``    Returns`` followed by the line containing ``    -------``.
    The parameter descriptions must start with the line containing parameter name and type:
    ``   parameter_name : type``. Parameter name may contain symbols '_', 'A-Z', 'a-z', '0'-'9' and
    start with a letter or '_'. Every line in parameter description section must be empty or
    start with FOUR spaces. The lines that contain description text may contain additional spaces.

    This function:

    - finds the parameter description section;

    - removes 4 spaces from the beginning of each line;

    - finds parameter names and descriptions for each parameter.

    Parameters
    ----------

    doc_string : str
        doc string as return by ``some_function.__doc__``. Every line of the docstring must be
        indented at least by 4 spaces.

    search_param_section : bool
        tells the function to separate the parameter section of the docstring before parsing it.
        If set ``False``, the function assumes that ``doc_string`` is contains only descriptions
        of parameters (already separated parameter section of the docstring). This option may
        be useful if the ``doc_string`` is not actual function docstring, but just the prepared
        description of parameters to be saved in the YAML file. Such description should not
        contain ``Parameters`` and ``Returns`` section titles. Note, that every line of the
        description must be indented at least by 4 spaces (as in docstrings).

    Returns
    -------

        A list of tuples ``(parameter_name, parameter_description)``:

        - ``parameter_name`` is a string

        - ``parameter_description`` is a list of strings
    """

    str_list = doc_string.split("\n")

    # Remove all spaces at the end of the strings (the should be no spaces there, but still)
    str_list = [s.rstrip() for s in str_list]

    if search_param_section:
        # We are interested only in the part of the docstring that contains description of parameters
        #   Google-style docstrings are expected
        n_first, n_last = None, None
        for n in range(1, len(str_list) - 1):
            if (str_list[n - 1] == "    Parameters") and re.search(r"^    -+$", str_list[n]):
                n_first = n + 1
            if (str_list[n] == "    Returns") and re.search(r"^    -+$", str_list[n + 1]):
                n_last = n - 1
                break
    else:
        # Not search for the parameter section. All the lines of the list should be parsed
        n_first, n_last = 0, len(str_list) - 1

    assert (n_first is not None) or (
        n_last is not None
    ), "Incorrect docstring format: 'Parameters' or 'Return' statement was not found in the docstring"

    # The list of strings contains parameter descriptions
    str_list = str_list[n_first : n_last + 1]
    # Each line must start with 4 spaces or be empty. Verify this
    assert all(
        [(not s) or re.search(r"^    ", s) for s in str_list]
    ), "Incorrect docstring format: parameter descriptions should be indented by at least FOUR spaces"
    # Now remove the spaces from nonempty lines
    str_list = [s[4:] if s else s for s in str_list]

    param_pos = [n for n, s in enumerate(str_list) if re.search(r"^[_A-Za-z][_A-Za-z0-9]* :", s)]
    n_parameters = len(param_pos)
    assert n_parameters, "Incorrect docstring format: no parameters were found"

    param_pos.append(len(str_list))  # Having the last index will be helpful

    param_names = [re.search("[_A-Za-z][_A-Za-z0-9]*", str_list[param_pos[n]])[0] for n in range(n_parameters)]
    param_descriptions = [str_list[param_pos[n] : param_pos[n + 1]] for n in range(n_parameters)]

    # Remove empty strings from the end of each description (if any)
    for pd in param_descriptions:
        while pd and (not pd[-1]):
            pd.pop(-1)

    # Check if some of the parameter has no descriptions (the number of line must be > 1)
    #   The fist line of the description is actually the
    assert all(
        [len(s) > 1 for s in param_descriptions]
    ), "Incomplete docstring: some parameters have not descriptions"

    params = list(zip(param_names, param_descriptions))

    return params


def _verify_parsed_docstring(parameters, param_dict):
    """
    Verification of consistency of the parameters extracted from docstring and
    dictionary of parameters (probably default parameters). There must be one-to-one
    match between the entries of the list of the parameters extracted from the docstring
    and the parameter dictionary.

    Parameters
    ----------

    parameters : list(tuple)
        The list of ``(parameter_name, parameter_description)`` tuples.
        Descriptions are not analyzed by this function.

    param_dict : dict
        The dictionary of parameters: key - parameter name, value - default value.
        Values are not analyized by this function.

    Returns
    -------

    Raises exception if there is mismatch between the parameter sets.
    """

    pd = param_dict.copy()
    missing_param_list = []
    for p_name, _ in parameters:
        if p_name in pd:
            del pd[p_name]
        else:
            missing_param_list.append(p_name)

    extra_param_list = pd.keys()

    err_msg_list = []
    if missing_param_list:
        err_msg_list.append(f"parameters that are in docstring, but not in the dictionary {missing_param_list}")
    if extra_param_list:
        err_msg_list.append(f"dictionary parameters that are not found in the docstring: {extra_param_list}")

    if err_msg_list:
        err_msg = "Parsed parameter verification failed: parameter set is inconsistent:\n"
        err_msg += ", ".join(err_msg_list)
        assert False, err_msg


_user_instructions_on_editing_yaml = """
--------------------------------------------------
Brief instructions on editing YAML parameter files
--------------------------------------------------

Parameters are presented in the form:

parameter_name: parameter_value

Parameter values may be numbers (integers or floating point numbers), strings, lists and
  dictionaries. STRINGS are presented without quotes:

parameter_name: some_string_value

or

parameter_name: Some string value

LISTS are presented in the form:

parameter_name:
- item1   # Items have the same indentation as 'parameter_name'
- item2
- item3

DICTIONARIES are presented in the form:

parameter_name:
  dict_key1: dict_value1  # Indented by 2 spaces
  dict_key2: dict_value2
  dict_key3:    # The value for this dictionary pair is an array with 3 items
  - list_item1  # Same indentation as 'dict_key3'
  - list_item2
  - list_item3

Some parameters may be set to the value of 'None'. Python value 'None' is represented
as 'null' in YAML files:

parameter_name: null  # The parameter has a value of None
--------------------------------------------------
"""


def create_yaml_parameter_file(
    *,
    file_path,
    function_docstring,
    param_value_dict,
    dir_create=False,
    file_overwrite=False,
    search_param_section=True,
    user_editing_instructions=_user_instructions_on_editing_yaml,
):
    """
    Creates YAML parameter file based on parameter names and descriptions from ``parameters``
    and default values from ``param_value_dict``. The file is supposed to have simple
    human-readable and editable structure, so the descriptions from the docstring are
    inserted above each parameter entry in the YAML file. Values from the ``param_value_dict``
    are used as the default values of the parameters. In addition, a set of instructions for the
    user may be inserted at the beginning of the YAML file (as comments). By default, the
    basic instructions on editing YAML file are inserted.

    Instead of Google-style docstring (which contains Parameters and Returns) statements,
    a string that contains only parameter descriptions may be supplied as ``function_docstring``.
    The parameter descriptions must still be formatted according to Google style and have
    indentation of at least 4 spaces. In this case search of the parameter section should
    be disabled by setting ``search_param_section=False``.

    The function should be used to create YAML file with default parameter values that are later
    modified by users according to their needs.

    Parameters
    ----------

    file_path : str
        absolute or relative path of the new YAML parameter file. The file may have any
        extension. The function may create the nonexisting directory or overwrite existing
        files if ``dir_create`` and/or ``file_overwrite`` are set ``True``.

    function_docstring : str
        doc string as return by ``some_function.__doc__``. Every line of the docstring must be
        indented at least by 4 spaces.

    param_value_dict : dict
        the dictionary that contains (param_name: param_value) pairs. The values of the parameters
        are saved in the YAML file (may be considered default values). ``function_docstring`` and
        ``param_value_dict`` must contain the same parameters (one-to-one match). Exception will
        be raised if there is mismatch between the number or names of parameters.

    dir_create : bool
        if set True, then the directory will be automatically created if it does not exist,
        if set False, then the exception will be raised if the directory does not exist

    file_overwrite : bool
        if set True, then existing parameter file may be overwritten with the new file if it exists,
        otherwise the exception ``IOError`` is be raised if file with the name ``file_path``
        already exists.

    search_param_section : bool
        tells the function to separate the parameter section of the docstring before parsing it.
        If set ``False``, the function assumes that ``doc_string`` is contains only descriptions
        of parameters (already separated parameter section of the docstring). This option may
        be useful if the ``doc_string`` is not actual function docstring, but just the prepared
        description of parameters to be saved in the YAML file. Such description should not
        contain ``Parameters`` and ``Returns`` section titles. Note, that every line of the
        description must be indented at least by 4 spaces (as in docstrings).

    user_editing_instructions : str
        text that is added to the beginning of the YAML file. Typically the text will contain
        the instruction for the user. The default set of instructions contain brief information
        on editing the YAML file. Before the text is added to the YAML file, the "# "
        is added to the beginning of each line, so that the text appears as a comment in
        the YAML file. Set to ``None`` if no instructions need to be added.

    Returns
    -------

    No value is returned

    The function will raise ``IOError`` if the directory in the ``file_path`` does not exist or
    the file already exists, unless the parameters ``dir_create`` and/or ``file_overwrite`` are
    set ``True``. The function will raise ``AssertionError`` if parsing of the docstring is
    not successful or there is a mismatch between parameter set found in docstring and
    the dictionary ``param_value_dict``.
    """

    # Convert to absolute path
    file_path = os.path.expanduser(file_path)
    file_path = os.path.abspath(file_path)

    # Parse docstring (returns the list of param_name/param_description pairs, may raise AssertionError)
    parameters = _parse_docstring_parameters(function_docstring, search_param_section=search_param_section)

    # Check if entries in 'parameters' and 'param_value_dict' match (may raise AssertionError).
    _verify_parsed_docstring(parameters, param_value_dict)

    # Check if file already exists
    file_exists = False
    if not file_overwrite:
        file_exists = os.path.exists(file_path)
    else:
        # We would like to overwrite only files, not directories etc.
        file_exists = os.path.exists(file_path) and not os.path.isfile(file_path)

    if file_exists:
        raise IOError(f"File '{file_path}' already exists")

    # Create the directory if necessary
    dir_path, _ = os.path.split(file_path)
    if not dir_create and not os.path.isdir(dir_path):
        raise IOError(f"Directory '{dir_path}' does not exist")

    # Attempt to create the directory
    os.makedirs(dir_path, exist_ok=True)

    # Create the file output
    s_output = ""
    s_output += (
        "# This file is autogenerated with the default parameters and expected to be modified by the user\n"
    )

    # Insert the user instructions if provided
    if user_editing_instructions:
        # Remove the last '\n'
        if user_editing_instructions[-1] == "\n":
            user_editing_instructions = user_editing_instructions[:-1]
        # Now add "# " to the beginning of each non-empty line,
        #   and replace empty lines with "#"
        list_lines = user_editing_instructions.split("\n")
        list_lines = [f"# {s}" if s else "#" for s in list_lines]
        user_editing_instructions = "\n".join(list_lines)
        s_output += user_editing_instructions + "\n\n"

    for p_name, p_desc in parameters:
        desc = [f"#  {p}" for p in p_desc]
        s = "\n".join(desc)
        s += "\n\n"

        # Print the dictionary entry itself: yaml.dump performs necessary formatting
        d = {p_name: param_value_dict[p_name]}
        s += yaml.dump(d, indent=4)
        s += "\n\n"
        s_output += s

    with open(file_path, "w") as f:
        f.write(s_output)


def read_yaml_parameter_file(*, file_path):
    """
    Reads YAML parameter file. It is assumed that the file is created by the function
    ``create_yaml_parameter_file`` and subsequent editing did not change the overall file
    layout. The function does not check the schema, so any compatible parameter file
    will be loaded as a dictionary.

    Parameters
    ----------

    file_path : str
        relative or absolute path to the YAML file. The exception ``IOError`` will be
        raised if the file does not exist.

    Returns
    -------

    The dictionary that contains (param_name, param_value) pairs
    """

    # Convert to absolute path
    file_path = os.path.expanduser(file_path)
    file_path = os.path.abspath(file_path)

    if not os.path.isfile(file_path):
        raise IOError(f"File '{file_path}' does not exist")

    with open(file_path, "r") as f:
        param_dict = yaml.load(f, Loader=yaml.FullLoader)

    return param_dict



import webbrowser
from datetime import datetime

from qtpy.QtWidgets import (
    QMainWindow,
    QMessageBox,
    QLabel,
    QAction,
    QDialog,
    QVBoxLayout,
    QDialogButtonBox,
    QHBoxLayout,
    QProgressBar,
)
from qtpy.QtCore import Qt, Slot, Signal
from qtpy.QtGui import QGuiApplication, QCursor

from .central_widget import TwoPanelWidget
from .useful_widgets import global_gui_variables
from .wnd_manage_emission_lines import WndManageEmissionLines
from .wnd_compute_roi_maps import WndComputeRoiMaps
from .wnd_load_quant_calibration import WndLoadQuantitativeCalibration
from .wnd_image_wizard import WndImageWizard
from .wnd_general_settings_for_fitting import WndGeneralFittingSettings
from .wnd_detailed_fitting_params import WndDetailedFittingParamsLines, WndDetailedFittingParamsShared

import pyxrf

import logging

logger = logging.getLogger(__name__)

_main_window_geometry = {
    "initial_height": 700,
    "initial_width": 1200,
    "min_height": 700,
    "min_width": 1200,
}


class MainWindow(QMainWindow):

    signal_fitting_parameters_changed = Signal()

    def __init__(self, *, gpc):
        """
        Parameters
        ----------
        gpc: object
            reference to a class that holds references to processing classes.
        """
        super().__init__()

        self._cursor_set = False  # Indicates if temporary 'wait' cursor is set

        self.gpc = gpc
        self.gui_vars = global_gui_variables
        self.gui_vars["ref_main_window"] = self

        self.wnd_manage_emission_lines = WndManageEmissionLines(gpc=self.gpc, gui_vars=self.gui_vars)
        self.wnd_compute_roi_maps = WndComputeRoiMaps(gpc=self.gpc, gui_vars=self.gui_vars)
        self.wnd_image_wizard = WndImageWizard(gpc=self.gpc, gui_vars=self.gui_vars)
        self.wnd_load_quantitative_calibration = WndLoadQuantitativeCalibration(
            gpc=self.gpc, gui_vars=self.gui_vars
        )
        self.wnd_general_fitting_settings = WndGeneralFittingSettings(gpc=self.gpc, gui_vars=self.gui_vars)
        self.wnd_fitting_parameters_shared = WndDetailedFittingParamsShared(gpc=self.gpc, gui_vars=self.gui_vars)
        self.wnd_fitting_parameters_lines = WndDetailedFittingParamsLines(gpc=self.gpc, gui_vars=self.gui_vars)
        # Indicates that the window was closed (used mostly for testing)
        self._is_closed = False

        global_gui_variables["gui_state"]["databroker_available"] = self.gpc.is_databroker_available()

        self.initialize()

        self.central_widget.left_panel.load_data_widget.update_main_window_title.connect(self.update_window_title)

        # Set the callback for update forms with fitting parameters
        def update_fitting_parameter_forms():
            self.signal_fitting_parameters_changed.emit()

        gpc.add_parameters_changed_cb(update_fitting_parameter_forms)
        gpc.param_model.parameters_changed()

    def initialize(self):

        self.resize(_main_window_geometry["initial_width"], _main_window_geometry["initial_height"])

        self.setMinimumWidth(_main_window_geometry["min_width"])
        self.setMinimumHeight(_main_window_geometry["min_height"])

        self.setWindowTitle(self.gpc.get_window_title())

        self.central_widget = TwoPanelWidget(gpc=self.gpc, gui_vars=self.gui_vars)
        self.setCentralWidget(self.central_widget)

        # Status bar
        self.statusLabel = QLabel()
        self.statusBar().addWidget(self.statusLabel)
        self.statusProgressBar = QProgressBar()
        self.statusProgressBar.setFixedWidth(200)
        self.statusBar().addPermanentWidget(self.statusProgressBar)

        self.statusLabelDefaultText = "No data is loaded"
        self.statusLabel.setText(self.statusLabelDefaultText)

        # 'Scan Data' menu item
        self.action_read_file = QAction("&Read File...", self)
        self.action_read_file.setStatusTip("Load data from HDF5 file")
        self.action_read_file.triggered.connect(self.central_widget.left_panel.load_data_widget.pb_file.clicked)

        self.action_load_run = QAction("&Load Run...", self)
        self.action_load_run.setEnabled(self.gui_vars["gui_state"]["databroker_available"])
        self.action_load_run.setStatusTip("Load data from database (Databroker)")
        self.action_load_run.triggered.connect(self.central_widget.left_panel.load_data_widget.pb_dbase.clicked)

        self.action_view_metadata = QAction("View Metadata...", self)
        self.action_view_metadata.setEnabled(self.gpc.is_scan_metadata_available())
        self.action_view_metadata.setStatusTip("View metadata for loaded run")
        self.action_view_metadata.triggered.connect(
            self.central_widget.left_panel.load_data_widget.pb_view_metadata.clicked
        )

        # Main menu
        menubar = self.menuBar()
        # Disable native menu bar (it doesn't work on MacOS 10.15 with PyQt<=5.11)
        #   It may work with later versions of PyQt when they become available.
        menubar.setNativeMenuBar(False)
        loadData = menubar.addMenu("Scan &Data")
        loadData.addAction(self.action_read_file)
        loadData.addAction(self.action_load_run)
        loadData.addSeparator()
        loadData.addAction(self.action_view_metadata)

        # 'Fitting Model' menu item
        self.action_lines_find_automatically = QAction("Find &Automatically...", self)
        self.action_lines_find_automatically.setStatusTip("Automatically find emission lines in total spectrum")
        self.action_lines_find_automatically.triggered.connect(
            self.central_widget.left_panel.model_widget.pb_find_elines.clicked
        )

        self.action_lines_load_from_file = QAction("Load From &File...", self)
        self.action_lines_load_from_file.setStatusTip(
            "Load processing parameters, including selected emission lines, from JSON file"
        )
        self.action_lines_load_from_file.triggered.connect(
            self.central_widget.left_panel.model_widget.pb_load_elines.clicked
        )

        self.action_lines_load_quant_standard = QAction("Load &Quantitative Standards...", self)
        self.action_lines_load_quant_standard.setStatusTip(
            "Load quantitative standard. The emission lines from the standard are automatically selected"
        )
        self.action_lines_load_quant_standard.triggered.connect(
            self.central_widget.left_panel.model_widget.pb_load_qstandard.clicked
        )

        self.action_add_remove_emission_lines = QAction("&Add/Remove Emission Lines...", self)
        self.action_add_remove_emission_lines.setStatusTip("Manually add and remove emission lines")
        self.action_add_remove_emission_lines.triggered.connect(
            self.central_widget.left_panel.model_widget.pb_manage_emission_lines.clicked
        )

        self.action_save_model_params = QAction("&Save Model Parameters...", self)
        self.action_save_model_params.setStatusTip("Save model parameters to JSON file")
        self.action_save_model_params.triggered.connect(
            self.central_widget.left_panel.model_widget.pb_save_elines.clicked
        )

        self.action_add_remove_emission_lines = QAction("Start Model &Fitting", self)
        self.action_add_remove_emission_lines.setStatusTip("Run computations: start fitting for total spectrum")
        self.action_add_remove_emission_lines.triggered.connect(
            self.central_widget.left_panel.model_widget.pb_start_fitting.clicked
        )

        fittingModel = menubar.addMenu("Fitting &Model")
        emissionLines = fittingModel.addMenu("&Emission Lines")
        emissionLines.addAction(self.action_lines_find_automatically)
        emissionLines.addAction(self.action_lines_load_from_file)
        emissionLines.addAction(self.action_lines_load_quant_standard)
        fittingModel.addAction(self.action_add_remove_emission_lines)
        fittingModel.addSeparator()
        fittingModel.addAction(self.action_save_model_params)
        fittingModel.addSeparator()
        fittingModel.addAction(self.action_add_remove_emission_lines)

        # "XRF Maps" menu item
        self.action_start_xrf_map_fitting = QAction("Start XRF Map &Fitting", self)
        self.action_start_xrf_map_fitting.setStatusTip("Run computations: start fitting for XRF maps")
        self.action_start_xrf_map_fitting.triggered.connect(
            self.central_widget.left_panel.fit_maps_widget.pb_start_map_fitting.clicked
        )

        self.action_compute_rois = QAction("Compute &ROIs...", self)
        self.action_compute_rois.setStatusTip("Compute XRF Maps based on spectral ROIs")
        self.action_compute_rois.triggered.connect(
            self.central_widget.left_panel.fit_maps_widget.pb_compute_roi_maps.clicked
        )

        self.action_load_quant_calibration = QAction("&Load Quantitative Calibration...", self)
        self.action_load_quant_calibration.setStatusTip(
            "Load quantitative calibration from JSON file. Calibration is used for scaling of XRF Maps"
        )
        self.action_load_quant_calibration.triggered.connect(
            self.central_widget.left_panel.fit_maps_widget.pb_load_quant_calib.clicked
        )

        self.action_save_quant_calibration = QAction("&Save Quantitative Calibration...", self)
        self.action_save_quant_calibration.setStatusTip(
            "Save Quantitative Calibration based on XRF map of the standard sample"
        )
        self.action_save_quant_calibration.triggered.connect(
            self.central_widget.left_panel.fit_maps_widget.pb_save_q_calibration.clicked
        )

        self.action_export_to_tiff_and_txt = QAction("&Export to TIFF and TXT...", self)
        self.action_export_to_tiff_and_txt.setStatusTip("Export XRF Maps as TIFF and/or TXT files")
        self.action_export_to_tiff_and_txt.triggered.connect(
            self.central_widget.left_panel.fit_maps_widget.pb_export_to_tiff_and_txt.clicked
        )

        xrfMaps = menubar.addMenu("XRF &Maps")
        xrfMaps.addAction(self.action_start_xrf_map_fitting)
        xrfMaps.addAction(self.action_compute_rois)
        xrfMaps.addSeparator()
        xrfMaps.addAction(self.action_load_quant_calibration)
        xrfMaps.addSeparator()
        xrfMaps.addAction(self.action_save_quant_calibration)
        xrfMaps.addAction(self.action_export_to_tiff_and_txt)

        # "View" menu item
        self.action_show_matplotlib_toolbar = QAction("Show &Matplotlib Toolbar", self)
        self.action_show_matplotlib_toolbar.setCheckable(True)
        self.action_show_matplotlib_toolbar.setChecked(True)
        self.action_show_matplotlib_toolbar.setStatusTip("Show Matplotlib Toolbar on the plots")
        self.action_show_matplotlib_toolbar.toggled.connect(self.action_show_matplotlib_toolbar_toggled)

        self.action_show_widget_tooltips = QAction("Show &Tooltips", self)
        self.action_show_widget_tooltips.setCheckable(True)
        self.action_show_widget_tooltips.setChecked(self.gui_vars["show_tooltip"])
        self.action_show_widget_tooltips.setStatusTip("Show widget tooltips")
        self.action_show_widget_tooltips.toggled.connect(self.action_show_widget_tooltips_toggled)

        options = menubar.addMenu("&Options")
        options.addAction(self.action_show_widget_tooltips)
        options.addAction(self.action_show_matplotlib_toolbar)

        # "Help" menu item
        self.action_online_docs = QAction("Online &Documentation", self)
        self.action_online_docs.setStatusTip("Open online documentation in the default browser")
        self.action_online_docs.triggered.connect(self.action_online_docs_triggered)

        self.action_about = QAction("&About PyXRF", self)
        self.action_about.setStatusTip("Show information about this program")
        self.action_about.triggered.connect(self.action_about_triggered)

        help = menubar.addMenu("&Help")
        help.addAction(self.action_online_docs)
        help.addSeparator()
        help.addAction(self.action_about)

        self.update_widget_state()

        # Connect signals
        self.central_widget.left_panel.load_data_widget.update_preview_map_range.connect(
            self.central_widget.right_panel.tab_preview_plots.preview_plot_count.update_map_range
        )

        # Before loading a new file or run
        self.central_widget.left_panel.load_data_widget.signal_loading_new_run.connect(
            self.central_widget.right_panel.slot_activate_tab_preview
        )

        # Open a new file or run
        self.central_widget.left_panel.load_data_widget.signal_new_run_loaded.connect(
            self.central_widget.left_panel.slot_activate_load_data_tab
        )
        self.central_widget.left_panel.load_data_widget.signal_new_run_loaded.connect(
            self.central_widget.right_panel.slot_activate_tab_preview
        )
        self.central_widget.left_panel.load_data_widget.signal_new_run_loaded.connect(self.slot_new_run_loaded)
        self.central_widget.left_panel.load_data_widget.signal_new_run_loaded.connect(
            self.wnd_image_wizard.slot_update_table
        )
        self.central_widget.left_panel.load_data_widget.signal_new_run_loaded.connect(
            self.central_widget.right_panel.tab_plot_xrf_maps.slot_update_dataset_info
        )
        self.central_widget.left_panel.load_data_widget.signal_new_run_loaded.connect(
            self.central_widget.right_panel.tab_plot_rgb_maps.slot_update_dataset_info
        )
        self.central_widget.left_panel.load_data_widget.signal_new_run_loaded.connect(
            self.wnd_load_quantitative_calibration.update_all_data
        )
        self.central_widget.left_panel.load_data_widget.signal_new_run_loaded.connect(
            self.central_widget.left_panel.fit_maps_widget.slot_update_for_new_loaded_run
        )

        # New model is loaded or processing parameters (incident energy) was changed
        self.central_widget.left_panel.model_widget.signal_incident_energy_or_range_changed.connect(
            self.central_widget.right_panel.tab_preview_plots.preview_plot_spectrum.redraw_preview_plot
        )
        self.central_widget.left_panel.model_widget.signal_incident_energy_or_range_changed.connect(
            self.wnd_manage_emission_lines.update_widget_data
        )
        self.central_widget.left_panel.model_widget.signal_incident_energy_or_range_changed.connect(
            self.central_widget.right_panel.tab_plot_fitting_model.redraw_plot_fit
        )

        self.central_widget.left_panel.model_widget.signal_model_loaded.connect(
            self.central_widget.right_panel.tab_preview_plots.preview_plot_spectrum.redraw_preview_plot
        )
        self.central_widget.left_panel.model_widget.signal_model_loaded.connect(
            self.central_widget.right_panel.slot_activate_tab_fitting_model
        )
        self.central_widget.left_panel.model_widget.signal_model_loaded.connect(
            self.central_widget.right_panel.tab_plot_fitting_model.update_controls
        )
        self.central_widget.left_panel.model_widget.signal_model_loaded.connect(
            self.wnd_manage_emission_lines.update_widget_data
        )

        # XRF Maps dataset changed
        self.central_widget.right_panel.tab_plot_xrf_maps.signal_maps_dataset_selection_changed.connect(
            self.wnd_image_wizard.slot_update_table
        )

        self.central_widget.right_panel.tab_plot_xrf_maps.signal_maps_dataset_selection_changed.connect(
            self.central_widget.right_panel.tab_plot_rgb_maps.combo_select_dataset_update_current_index
        )
        self.central_widget.right_panel.tab_plot_rgb_maps.signal_rgb_maps_dataset_selection_changed.connect(
            self.central_widget.right_panel.tab_plot_xrf_maps.combo_select_dataset_update_current_index
        )

        self.central_widget.right_panel.tab_plot_xrf_maps.signal_maps_norm_changed.connect(
            self.wnd_image_wizard.slot_update_ranges
        )

        # Quantitative calibration changed
        self.wnd_load_quantitative_calibration.signal_quantitative_calibration_changed.connect(
            self.central_widget.right_panel.tab_plot_rgb_maps.slot_update_ranges
        )
        self.wnd_load_quantitative_calibration.signal_quantitative_calibration_changed.connect(
            self.wnd_image_wizard.slot_update_ranges
        )

        # Selected element is changed (tools for emission line selection)
        self.wnd_manage_emission_lines.signal_selected_element_changed.connect(
            self.central_widget.right_panel.tab_plot_fitting_model.slot_selection_item_changed
        )
        self.central_widget.right_panel.tab_plot_fitting_model.signal_selected_element_changed.connect(
            self.wnd_manage_emission_lines.slot_selection_item_changed
        )
        self.central_widget.right_panel.tab_plot_fitting_model.signal_add_line.connect(
            self.wnd_manage_emission_lines.pb_add_eline_clicked
        )
        self.central_widget.right_panel.tab_plot_fitting_model.signal_remove_line.connect(
            self.wnd_manage_emission_lines.pb_remove_eline_clicked
        )
        self.wnd_manage_emission_lines.signal_update_element_selection_list.connect(
            self.central_widget.right_panel.tab_plot_fitting_model.slot_update_eline_selection_list
        )
        self.wnd_manage_emission_lines.signal_update_add_remove_btn_state.connect(
            self.central_widget.right_panel.tab_plot_fitting_model.slot_update_add_remove_btn_state
        )

        self.wnd_manage_emission_lines.signal_selected_element_changed.connect(
            self.central_widget.left_panel.model_widget.slot_selection_item_changed
        )
        self.central_widget.right_panel.tab_plot_fitting_model.signal_selected_element_changed.connect(
            self.central_widget.left_panel.model_widget.slot_selection_item_changed
        )

        # Total spectrum fitting completed
        self.central_widget.left_panel.model_widget.signal_total_spectrum_fitting_completed.connect(
            self.wnd_manage_emission_lines.update_eline_table
        )
        # Total spectrum invalidated
        self.wnd_manage_emission_lines.signal_parameters_changed.connect(
            self.central_widget.left_panel.model_widget.update_fit_status
        )
        # New dataset loaded or different channel selected. Compute fit parameters.
        self.central_widget.left_panel.load_data_widget.signal_data_channel_changed.connect(
            self.central_widget.left_panel.model_widget.clear_fit_status
        )
        self.central_widget.left_panel.load_data_widget.signal_new_run_loaded.connect(
            self.central_widget.left_panel.model_widget.clear_fit_status
        )
        self.central_widget.left_panel.model_widget.signal_incident_energy_or_range_changed.connect(
            self.central_widget.left_panel.model_widget.clear_fit_status
        )

        # Update map datasets (Fitted maps)
        self.central_widget.left_panel.fit_maps_widget.signal_map_fitting_complete.connect(
            self.central_widget.right_panel.tab_plot_xrf_maps.slot_update_dataset_info
        )
        self.central_widget.left_panel.fit_maps_widget.signal_map_fitting_complete.connect(
            self.central_widget.right_panel.tab_plot_rgb_maps.slot_update_dataset_info
        )
        self.central_widget.left_panel.fit_maps_widget.signal_activate_tab_xrf_maps.connect(
            self.central_widget.right_panel.slot_activate_tab_xrf_maps
        )

        # Update map datasets (ROI maps)
        self.wnd_compute_roi_maps.signal_roi_computation_complete.connect(
            self.central_widget.right_panel.tab_plot_xrf_maps.slot_update_dataset_info
        )
        self.wnd_compute_roi_maps.signal_roi_computation_complete.connect(
            self.central_widget.right_panel.tab_plot_rgb_maps.slot_update_dataset_info
        )
        self.wnd_compute_roi_maps.signal_activate_tab_xrf_maps.connect(
            self.central_widget.right_panel.slot_activate_tab_xrf_maps
        )

        self.signal_fitting_parameters_changed.connect(self.wnd_general_fitting_settings.update_form_data)
        self.signal_fitting_parameters_changed.connect(self.wnd_fitting_parameters_shared.update_form_data)
        self.signal_fitting_parameters_changed.connect(self.wnd_fitting_parameters_lines.update_form_data)
        self.signal_fitting_parameters_changed.connect(self.wnd_manage_emission_lines.update_widget_data)

    @Slot()
    @Slot(str)
    def update_widget_state(self, condition=None):
        # Update the state of the menu bar
        state = not self.gui_vars["gui_state"]["running_computations"]
        self.menuBar().setEnabled(state)

        state_computations = self.gui_vars["gui_state"]["running_computations"]
        if state_computations:
            if not self._cursor_set:
                QGuiApplication.setOverrideCursor(QCursor(Qt.WaitCursor))
                self._cursor_set = True
        else:
            if self._cursor_set:
                QGuiApplication.restoreOverrideCursor()
                self._cursor_set = False

        # Forward to children
        self.central_widget.update_widget_state(condition)

        # Forward the updates to open windows
        self.wnd_manage_emission_lines.update_widget_state(condition)
        self.wnd_compute_roi_maps.update_widget_state(condition)
        self.wnd_image_wizard.update_widget_state(condition)
        self.wnd_load_quantitative_calibration.update_widget_state(condition)
        self.wnd_general_fitting_settings.update_widget_state(condition)
        self.wnd_fitting_parameters_shared.update_widget_state(condition)
        self.wnd_fitting_parameters_lines.update_widget_state(condition)

    def closeEvent(self, event):
        mb_close = QMessageBox(
            QMessageBox.Question,
            "Exit",
            "Are you sure you want to EXIT the program?",
            QMessageBox.Yes | QMessageBox.No,
            parent=self,
        )
        mb_close.setDefaultButton(QMessageBox.No)

        if mb_close.exec() == QMessageBox.Yes:
            event.accept()

            self.wnd_manage_emission_lines.close()
            self.wnd_compute_roi_maps.close()
            self.wnd_image_wizard.close()
            self.wnd_load_quantitative_calibration.close()
            self.wnd_general_fitting_settings.close()
            self.wnd_fitting_parameters_shared.close()
            self.wnd_fitting_parameters_lines.close()

            # This flag is used for CI tests
            self._is_closed = True
        else:
            event.ignore()

    def action_online_docs_triggered(self):
        """
        Display online documentation: open the URL in the default browser.
        """
        doc_url = "http://nsls-ii.github.io/PyXRF/"
        try:
            webbrowser.open(doc_url, autoraise=True)
        except Exception as ex:
            logger.error(f"Error occurred while opening URL '{doc_url}' in the default browser")
            msg = f"Failed to Open Online Documentation. \n  Exception: {str(ex)}"
            msgbox = QMessageBox(QMessageBox.Critical, "Error", msg, QMessageBox.Ok, parent=self)
            msgbox.exec()

    def action_about_triggered(self):
        """
        Display 'About' dialog box
        """
        dlg = DialogAbout()
        dlg.exec()

    def action_show_matplotlib_toolbar_toggled(self, state):
        """
        Turn tooltips on or off
        """
        self.gui_vars["show_matplotlib_toolbar"] = state
        self.update_widget_state()

    def action_show_widget_tooltips_toggled(self, state):
        """
        Turn tooltips on or off
        """
        self.gui_vars["show_tooltip"] = state
        self.update_widget_state("tooltips")

    @Slot()
    def update_window_title(self):
        self.setWindowTitle(self.gpc.get_window_title())

    @Slot(bool)
    def slot_new_run_loaded(self, success):
        if success:
            # Update status bar
            file_name = self.gpc.get_loaded_file_name()
            run_id, run_uid = "", ""
            if self.gpc.is_scan_metadata_available():
                try:
                    run_id = self.gpc.get_metadata_scan_id()
                except Exception:
                    pass
                try:
                    run_uid = self.gpc.get_metadata_scan_uid()
                except Exception:
                    pass
            else:
                if self.gpc.get_current_run_id() >= 0:
                    run_id = self.gpc.get_current_run_id()
            s = ""
            if run_id:
                s += f"ID: {run_id}  "
            if run_uid:
                if run_id:
                    s += f"({run_uid})  "
                else:
                    s += f"UID: {run_uid}  "
            if file_name:
                s += f"File: '{file_name}'"
            self._set_status_bar_text(s)

            # Activate/deactivate "View Metadata ..." menu item
            if self.gpc.is_scan_metadata_available():
                self.action_view_metadata.setEnabled(True)
            else:
                self.action_view_metadata.setEnabled(False)

        else:
            self._set_status_bar_text()

    def _set_status_bar_text(self, text=None):
        if text is None:
            text = self.statusLabelDefaultText
        self.statusLabel.setText(text)


class DialogAbout(QDialog):
    def __init__(self):
        super().__init__()

        self.setWindowTitle("About PyXRF")
        self.setFixedSize(500, 500)

        text_name = "PyXRF"
        text_description = "X-Ray Fluorescence Analysis Tool"

        text_ver = f"Version: {pyxrf.__version__}"
        text_latest_ver = "Latest stable version:"

        text_credit = "Credits:"
        text_credit_org = (
            "Data Acquisition, Management and Analysis Group\n"
            "National Synchrontron Light Source II\n"
            "Brookhaven National Laboratory"
        )

        text_copyright = f"\u00A92015\u2014{datetime.now().year} Brookhaven National Laboratory"

        label_name = QLabel(text_name)
        label_name.setStyleSheet("QLabel {font-weight: bold; font-size: 32px}")

        label_description = QLabel(text_description)
        label_description.setStyleSheet("QLabel {font-style: italic; font-size: 18px}")

        label_ver = QLabel(text_ver)
        label_latest_ver = QLabel(text_latest_ver)
        label_credit = QLabel(text_credit)
        label_org = QLabel(text_credit_org)
        label_copyright = QLabel(text_copyright)

        button_box = QDialogButtonBox(QDialogButtonBox.Close)
        button_box.accepted.connect(self.accept)
        button_box.rejected.connect(self.reject)

        vbox = QVBoxLayout()

        vbox.addStretch(1)

        hbox = QHBoxLayout()
        hbox.addStretch(1)
        hbox.addWidget(label_name)
        hbox.addStretch(1)
        vbox.addLayout(hbox)

        hbox = QHBoxLayout()
        hbox.addStretch(1)
        hbox.addWidget(label_description)
        hbox.addStretch(1)
        vbox.addLayout(hbox)

        vbox.addStretch(1)

        hbox = QHBoxLayout()
        hbox.addSpacing(30)
        hbox.addWidget(label_ver)
        hbox.addStretch(1)
        vbox.addLayout(hbox)

        hbox = QHBoxLayout()
        hbox.addSpacing(30)
        hbox.addWidget(label_latest_ver)
        hbox.addStretch(1)
        vbox.addLayout(hbox)

        vbox.addStretch(1)

        hbox = QHBoxLayout()
        hbox.addSpacing(30)
        hbox.addWidget(label_credit, 0, Qt.AlignTop)
        hbox.addWidget(label_org, 0, Qt.AlignTop)
        hbox.addStretch(1)
        vbox.addLayout(hbox)

        vbox.addSpacing(20)

        hbox = QHBoxLayout()
        hbox.addSpacing(30)
        hbox.addWidget(label_copyright)
        hbox.addStretch(1)
        vbox.addLayout(hbox)

        vbox.addSpacing(20)

        vbox.addWidget(button_box)

        self.setLayout(vbox)



from qtpy.QtWidgets import QWidget, QHBoxLayout, QRadioButton, QButtonGroup, QTableWidget, QHeaderView, QSizePolicy
from qtpy.QtGui import QPalette
from qtpy.QtCore import Qt, Signal

import copy

from .useful_widgets import RangeManager, get_background_css, ComboBoxNamed

import logging

logger = logging.getLogger(__name__)


class RgbSelectionWidget(QWidget):

    signal_update_map_selections = Signal()

    def __init__(self):
        super().__init__()

        self._range_table = []
        self._limit_table = []
        self._rgb_keys = ["red", "green", "blue"]
        self._rgb_dict = {_: None for _ in self._rgb_keys}

        widget_layout = self._setup_rgb_widget()
        self.setLayout(widget_layout)

        sp = QSizePolicy()
        sp.setControlType(QSizePolicy.PushButton)
        sp.setHorizontalPolicy(QSizePolicy.Expanding)
        sp.setVerticalPolicy(QSizePolicy.Fixed)
        self.setSizePolicy(sp)

    def _setup_rgb_element(self, n_row, *, rb_check=0):
        """
        Parameters
        ----------
        rb_check: int
            The number of QRadioButton to check. Typically this would be the row number.
        """
        combo_elements = ComboBoxNamed(name=f"{n_row}")
        # combo_elements.setSizeAdjustPolicy(QComboBox.AdjustToContents)

        # Set text color for QComboBox widget (necessary if the program is used with Dark theme)
        pal = combo_elements.palette()
        pal.setColor(QPalette.ButtonText, Qt.black)
        combo_elements.setPalette(pal)
        # Set text color for drop-down view (necessary if the program is used with Dark theme)
        pal = combo_elements.view().palette()
        pal.setColor(QPalette.Text, Qt.black)
        combo_elements.view().setPalette(pal)

        btns = [QRadioButton(), QRadioButton(), QRadioButton()]
        if 0 <= rb_check < len(btns):
            btns[rb_check].setChecked(True)

        # Color is set for operation with Dark theme
        for btn in btns:
            pal = btn.palette()
            pal.setColor(QPalette.Text, Qt.black)
            btn.setPalette(pal)

        btn_group = QButtonGroup()
        for btn in btns:
            btn_group.addButton(btn)

        rng = RangeManager(name=f"{n_row}", add_sliders=True)
        rng.setTextColor([0, 0, 0])  # Set text color to 'black'
        # Set some text in edit boxes (just to demonstrate how the controls will look like)
        rng.le_min_value.setText("0.0")
        rng.le_max_value.setText("1.0")

        rng.setAlignment(Qt.AlignCenter)
        return combo_elements, btns, rng, btn_group

    def _enable_selection_events(self, enable):
        if enable:
            if not self.elements_btn_groups_events_enabled:
                for btn_group in self.elements_btn_groups:
                    btn_group.buttonToggled.connect(self.rb_toggled)
                for el_combo in self.elements_combo:
                    el_combo.currentIndexChanged.connect(self.combo_element_current_index_changed)
                for el_range in self.elements_range:
                    el_range.selection_changed.connect(self.range_selection_changed)
                self.elements_btn_groups_events_enabled = True
        else:
            if self.elements_btn_groups_events_enabled:
                for btn_group in self.elements_btn_groups:
                    btn_group.buttonToggled.disconnect(self.rb_toggled)
                for el_combo in self.elements_combo:
                    el_combo.currentIndexChanged.disconnect(self.combo_element_current_index_changed)
                # Disconnecting the Range Manager signals is not necessary, but let's do it for consistency
                for el_range in self.elements_range:
                    el_range.selection_changed.disconnect(self.range_selection_changed)
                self.elements_btn_groups_events_enabled = False

    def _setup_rgb_widget(self):

        self.elements_combo = []
        self.elements_rb_color = []
        self.elements_range = []
        self.elements_btn_groups = []
        self.elements_btn_groups_events_enabled = False
        self.row_colors = []

        self.table = QTableWidget()
        # Horizontal header entries
        tbl_labels = ["Element", "Red", "Green", "Blue", "Range"]
        # The list of columns that stretch with the table
        self.tbl_cols_stretch = ("Range",)

        self.table.setColumnCount(len(tbl_labels))
        self.table.setRowCount(3)
        self.table.setHorizontalHeaderLabels(tbl_labels)
        self.table.verticalHeader().hide()
        self.table.setSelectionMode(QTableWidget.NoSelection)

        header = self.table.horizontalHeader()
        for n, lbl in enumerate(tbl_labels):
            # Set stretching for the columns
            if lbl in self.tbl_cols_stretch:
                header.setSectionResizeMode(n, QHeaderView.Stretch)
            else:
                header.setSectionResizeMode(n, QHeaderView.ResizeToContents)

        vheader = self.table.verticalHeader()
        vheader.setSectionResizeMode(QHeaderView.Stretch)  # ResizeToContents)

        for n_row in range(3):
            combo_elements, btns, rng, btn_group = self._setup_rgb_element(n_row, rb_check=n_row)

            combo_elements.setMinimumWidth(180)
            self.table.setCellWidget(n_row, 0, combo_elements)
            for i, btn in enumerate(btns):
                item = QWidget()
                item_hbox = QHBoxLayout(item)
                item_hbox.addWidget(btn)
                item_hbox.setAlignment(Qt.AlignCenter)
                item_hbox.setContentsMargins(0, 0, 0, 0)
                item.setMinimumWidth(70)

                self.table.setCellWidget(n_row, i + 1, item)

            rng.setMinimumWidth(200)
            rng.setMaximumWidth(400)
            self.table.setCellWidget(n_row, 4, rng)

            self.elements_combo.append(combo_elements)
            self.elements_rb_color.append(btns)
            self.elements_range.append(rng)
            self.elements_btn_groups.append(btn_group)
            self.row_colors.append(self._rgb_keys[n_row])

        # Colors that are used to paint rows of the table in RGB colors
        br = 150
        self._rgb_row_colors = {"red": (255, br, br), "green": (br, 255, br), "blue": (br, br, 255)}
        self._rgb_color_keys = ["red", "green", "blue"]

        # Set initial colors
        for n_row in range(self.table.rowCount()):
            self._set_row_color(n_row)

        self._enable_selection_events(True)

        self.table.resizeRowsToContents()

        # Table height is computed based on content. It doesn't seem
        #   to account for the height of custom widgets, but the table
        #   looks good enough
        table_height = 0
        for n_row in range(self.table.rowCount()):
            table_height += self.table.rowHeight(n_row)
        self.table.setMaximumHeight(table_height)

        table_width = 650
        self.table.setMinimumWidth(table_width)
        self.table.setMaximumWidth(800)

        hbox = QHBoxLayout()
        hbox.addWidget(self.table)

        return hbox

    def combo_element_current_index_changed(self, name, index):
        if index < 0 or index >= len(self._range_table):
            return
        n_row = int(name)
        sel_eline = self._range_table[index][0]
        row_color = self.row_colors[n_row]
        self._rgb_dict[row_color] = sel_eline

        self.elements_range[n_row].set_range(self._range_table[index][1], self._range_table[index][2])
        self.elements_range[n_row].set_selection(
            value_low=self._limit_table[index][1], value_high=self._limit_table[index][2]
        )
        self._update_map_selections()

    def range_selection_changed(self, v_low, v_high, name):
        n_row = int(name)
        row_color = self.row_colors[n_row]
        sel_eline = self._rgb_dict[row_color]
        ind = None
        try:
            ind = [_[0] for _ in self._limit_table].index(sel_eline)
        except ValueError:
            pass
        if ind is not None:
            self._limit_table[ind][1] = v_low
            self._limit_table[ind][2] = v_high
            self._update_map_selections()
        # We are not preventing users to select the same emission line in to rows.
        #   Update the selected limits in other rows where the same element is selected.
        for nr, el_range in enumerate(self.elements_range):
            if (nr != n_row) and (self._rgb_dict[self.row_colors[nr]] == sel_eline):
                el_range.set_selection(value_low=v_low, value_high=v_high)

    def _get_selected_row_color(self, n_row):

        color_key = None

        btns = self.elements_rb_color[n_row]

        for n, btn in enumerate(btns):
            if btn.isChecked():
                color_key = self._rgb_color_keys[n]
                break

        return color_key

    def _set_row_color(self, n_row, *, color_key=None):
        """
        Parameters
        ----------
        n_row: int
            The row number that needs background color change (0..2 if table has 3 rows)
        color_key: int
            Color key: "red", "green" or "blue"
        """

        if color_key is None:
            color_key = self._get_selected_row_color(n_row)
        if color_key is None:
            return

        self.row_colors[n_row] = color_key
        rgb = self._rgb_row_colors[color_key]

        # The following code is based on the arrangement of the widgets in the table
        #   Modify the code if widgets are arranged differently or the table structure
        #   is changed
        for n_col in range(self.table.columnCount()):
            wd = self.table.cellWidget(n_row, n_col)
            if n_col == 0:
                # Combo box: update both QComboBox and QWidget backgrounds
                #   QWidget - background of the drop-down selection list
                css1 = get_background_css(rgb, widget="QComboBox", editable=False)
                css2 = get_background_css(rgb, widget="QWidget", editable=True)
                wd.setStyleSheet(css2 + css1)
            elif n_col <= 3:
                # 3 QRadioButton's. The buttons are inserted into QWidget objects,
                #   and we need to change backgrounds of QWidgets, not only buttons.
                wd.setStyleSheet(get_background_css(rgb, widget="QWidget", editable=False))
            elif n_col == 4:
                # Custom RangeManager widget, color is updated using custom method
                wd.setBackground(rgb)

        n_col = self._rgb_color_keys.index(color_key)
        for n, n_btn in enumerate(self.elements_rb_color[n_row]):
            check_status = True if n == n_col else False
            n_btn.setChecked(check_status)

    def _fill_table(self):
        self._enable_selection_events(False)

        eline_list = [_[0] for _ in self._range_table]
        for n_row in range(self.table.rowCount()):
            self.elements_combo[n_row].clear()
            self.elements_combo[n_row].addItems(eline_list)

        for n_row, color in enumerate(self._rgb_color_keys):
            # Initially set colors in order
            self._set_row_color(n_row, color_key=color)
            eline_key = self._rgb_dict[color]
            if eline_key is not None:
                try:
                    ind = eline_list.index(eline_key)
                    self.elements_combo[n_row].setCurrentIndex(ind)
                    range_low, range_high = self._range_table[ind][1:]
                    self.elements_range[n_row].set_range(range_low, range_high)
                    sel_low, sel_high = self._limit_table[ind][1:]
                    self.elements_range[n_row].set_selection(value_low=sel_low, value_high=sel_high)
                except ValueError:
                    pass
            else:
                self.elements_combo[n_row].setCurrentIndex(-1)  # Deselect all
                self.elements_range[n_row].set_range(0, 1)
                self.elements_range[n_row].set_selection(value_low=0, value_high=1)

        self._enable_selection_events(True)

    def _find_rbutton(self, button):
        for nr, btns in enumerate(self.elements_rb_color):
            for nc, btn in enumerate(btns):
                if btn == button:
                    # Return tuple (nr, nc)
                    return nr, nc
        # Return None if the button is not found (this shouldn't happen)
        return None

    def rb_toggled(self, button, state):
        if state:  # Ignore signals from unchecked buttons
            nr, nc = self._find_rbutton(button)

            color_current = self.row_colors[nr]
            color_to_set = self._rgb_color_keys[nc]
            nr_switch = self.row_colors.index(color_to_set)

            self._enable_selection_events(False)

            self._set_row_color(nr, color_key=color_to_set)
            self._set_row_color(nr_switch, color_key=color_current)
            # Swap selected maps
            tmp = self._rgb_dict[color_to_set]
            self._rgb_dict[color_to_set] = self._rgb_dict[color_current]
            self._rgb_dict[color_current] = tmp

            self._enable_selection_events(True)

            self._update_map_selections()

    def set_ranges_and_limits(self, *, range_table=None, limit_table=None, rgb_dict=None):
        if range_table is not None:
            self._range_table = copy.deepcopy(range_table)
        if limit_table is not None:
            self._limit_table = copy.deepcopy(limit_table)
        if rgb_dict is not None:
            self._rgb_dict = rgb_dict.copy()
        self._fill_table()

    def _update_map_selections(self):
        """Upload the selections (limit table) and update plot"""
        self.signal_update_map_selections.emit()



import numpy as np
import scipy
import time as ttime

import logging

logger = logging.getLogger(__name__)


# =================================================================================
#  The following set of functions are separated from the rest of the program
#  and prepared to be moved to scikit-beam (skbeam.core.fitting.xrf_model)


def grid_interpolate(data, xx, yy, xx_uniform=None, yy_uniform=None):
    """
    Interpolate unevenly sampled data to even grid. The new even grid has the same
    dimensions as the original data and covers full range of original X and Y axes.

    Parameters
    ----------

    data : ndarray
        2D array with data values (`xx`, `yy` and `data` must have the same shape)
        ``data`` may be None. In this case interpolation will not be performed, but uniform
        grid will be generated. Use this feature to generate uniform grid.
    xx : ndarray
        2D array with measured values of X coordinates of data points (the values may be unevenly spaced)
    yy : ndarray
        2D array with measured values of Y coordinates of data points (the values may be unevenly spaced)
    xx_uniform : ndarray
        2D array with evenly spaced X axis values (same shape as `data`). If not provided, then
        generated automatically and returned by the function.
    yy_uniform : ndarray
        2D array with evenly spaced Y axis values (same shape as `data`). If not provided, then
        generated automatically and returned by the function.

    Returns
    -------
    data_uniform : ndarray
        2D array with data fitted to even grid (same shape as `data`)
    xx_uniform : ndarray
        2D array with evenly spaced X axis values (same shape as `data`)
    yy_uniform : ndarray
        2D array with evenly spaced Y axis values (same shape as `data`)
    """

    # Check if data shape and shape of coordinate arrays match
    if data is not None:
        if data.shape != xx.shape:
            msg = "Shapes of data and coordinate arrays do not match. (function 'grid_interpolate')"
            raise ValueError(msg)
    if xx.shape != yy.shape:
        msg = "Shapes of coordinate arrays 'xx' and 'yy' do not match. (function 'grid_interpolate')"
        raise ValueError(msg)
    if (xx_uniform is not None) and (xx_uniform.shape != xx.shape):
        msg = (
            "Shapes of data and array of uniform coordinates 'xx_uniform' do not match. "
            "(function 'grid_interpolate')"
        )
        raise ValueError(msg)
    if (yy_uniform is not None) and (xx_uniform.shape != xx.shape):
        msg = (
            "Shapes of data and array of uniform coordinates 'yy_uniform' do not match. "
            "(function 'grid_interpolate')"
        )
        raise ValueError(msg)

    ny, nx = xx.shape
    # Data must be 2-dimensional to use the following interpolation procedure.
    if (nx <= 1) or (ny <= 1):
        logger.debug("Function utils.grid_interpolate: single row or column scan. Grid interpolation is skipped")
        return data, xx, yy

    def _get_range(vv):
        """
        Returns the range of the data coordinates along X or Y axis. Coordinate
        data for a single axis is represented as a 2D array ``vv``. The array
        will have all rows or all columns identical or almost identical.
        The range is returned as ``vv_min`` (leftmost or topmost value)
        and ``vv_max`` (rightmost or bottommost value). Note, that ``vv_min`` may
        be greater than ``vv_max``

        Parameters
        ----------
        vv : ndarray
            2-d array of coordinates

        Returns
        -------
        vv_min : float
            starting point of the range
        vv_max : float
            end of the range
        """
        # The assumption is that X values are mostly changing along the dimension 1 and
        #   Y values change along the dimension 0 of the 2D array and only slightly change
        #   along the alternative dimension. Determine, if the range is for X or Y
        #   axis based on the dimension in which value change is the largest.
        if abs(vv[0, 0] - vv[0, -1]) > abs(vv[0, 0] - vv[-1, 0]):
            vv_min = np.median(vv[:, 0])
            vv_max = np.median(vv[:, -1])
        else:
            vv_min = np.median(vv[0, :])
            vv_max = np.median(vv[-1, :])

        return vv_min, vv_max

    if xx_uniform is None or yy_uniform is None:
        # Find the range of axes
        x_min, x_max = _get_range(xx)
        y_min, y_max = _get_range(yy)
        _yy_uniform, _xx_uniform = np.mgrid[y_min : y_max : ny * 1j, x_min : x_max : nx * 1j]

    if xx_uniform is None:
        xx_uniform = _xx_uniform
    if yy_uniform is None:
        yy_uniform = _yy_uniform

    xx = xx.flatten()
    yy = yy.flatten()
    xxyy = np.stack((xx, yy)).T

    if data is not None:
        # Do the interpolation only if data is provided
        data = data.flatten()
        # Do the interpolation
        data_uniform = scipy.interpolate.griddata(
            xxyy, data, (xx_uniform, yy_uniform), method="linear", fill_value=0
        )
    else:
        data_uniform = None

    return data_uniform, xx_uniform, yy_uniform


def normalize_data_by_scaler(data_in, scaler, *, data_name=None, name_not_scalable=None):
    """
    Normalize data based on the availability of scaler

    Parameters
    ----------

    data_in : ndarray
        numpy array of input data
    scaler : ndarray
        numpy array of scaling data, the same size as data_in
    data_name : str
        name of the data set ('time' or 'i0' etc.)
    name_not_scalable : list
        names of not scalable datasets (['time', 'i0_time'])

    Returns
    -------
    ndarray with normalized data, the same shape as data_in
        The returned array is the reference to 'data_in' if no normalization
        is applied to data or reference to modified copy of 'data_in' if
        normalization was applied.

    ::note::

        Normalization will not be performed if the following is true:

        - scaler is None

        - scaler is not the same shape as data_in

        - scaler contains all elements equal to zero

        If normalization is not performed then REFERENCE to data_in is returned.

    """

    if data_in is None or scaler is None:  # Nothing to scale
        logger.debug(
            "Function utils.normalize_data_by_scaler: data and/or scaler arrays are None. "
            "Data scaling is skipped."
        )
        return data_in

    if data_in.shape != scaler.shape:
        logger.debug(
            "Function utils.normalize_data_by_scaler: data and scaler arrays have different shape. "
            "Data scaling is skipped."
        )
        return data_in

    do_scaling = False
    # Check if data name is in the list of non-scalable items
    # If data name or the list does not exits, then do the scaling
    if name_not_scalable is None or data_name is None or data_name not in name_not_scalable:
        do_scaling = True

    # If scaler is all zeros, then don't scale the data:
    #   check if there is at least one nonzero element
    n_nonzero = np.count_nonzero(scaler)
    if not n_nonzero:
        logger.debug(
            "Function utils.normalize_data_by_scaler: scaler is all-zeros array. Data scaling is skipped."
        )
        do_scaling = False

    if do_scaling:
        # If scaler contains some zeros, set those zeros to mean value
        if data_in.size != n_nonzero:
            s_mean = np.mean(scaler[scaler != 0])
            # Avoid division by very small number (or zero)
            if np.abs(s_mean) < 1e-10:
                s_mean = 1e-10 if np.sign(s_mean) >= 0 else -1e-10
            scaler = scaler.copy()
            scaler[scaler == 0.0] = s_mean
        data_out = data_in / scaler
    else:
        data_out = data_in

    return data_out


# ===============================================================================
# The following functions are prepared to be moved to scikit-beam


def _get_2_sqrt_2_log2():
    return 2 * np.sqrt(2 * np.log(2))


def gaussian_sigma_to_fwhm(sigma):
    """
    Converts parameters of Gaussian curve: 'sigma' to 'fwhm'

    Parameters
    ----------

    sigma : float
        sigma of the Gaussian curve

    Returns
    -------
    FWHM of the Gaussian curve
    """
    return sigma * _get_2_sqrt_2_log2()


def gaussian_fwhm_to_sigma(fwhm):
    """
    Converts parameters of Gaussian curve: 'fwhm' to 'sigma'

    Parameters
    ----------

    fwhm : float
        Full Width at Half Maximum of the Gaussian curve

    Returns
    -------
    sigma of the Gaussian curve
    """
    return fwhm / _get_2_sqrt_2_log2()


def _get_sqrt_2_pi():
    return np.sqrt(2 * np.pi)


def gaussian_max_to_area(peak_max, peak_sigma):
    """
    Computes the area under Gaussian curve based on maximum and sigma

    Parameters
    ----------

    peak_max : float
        maximum of the Gaussian curve
    peak_sigma : float
        sigma of the Gaussian curve

    Returns
    -------
    area under the Gaussian curve
    """
    return peak_max * peak_sigma * _get_sqrt_2_pi()


def gaussian_area_to_max(peak_area, peak_sigma):
    """
    Computes the maximum of the Gaussian curve based on area
    under the curve and sigma

    Parameters
    ----------

    peak_area : float
       area under the Gaussian curve
    peak_sigma : float
        sigma of the Gaussian curve

    Returns
    -------
    area under the Gaussian curve
    """
    if peak_sigma == 0:
        return 0
    else:
        return peak_area / peak_sigma / _get_sqrt_2_pi()


# ==================================================================================


def convert_time_to_nexus_string(t):
    """
    Convert time to a string according to NEXUS format

    Parameters
    ----------

    t : time.struct_time
        Time in the format returned by ``time.localtime`` or ``time.gmtime``

    Returns
    -------

    t : str
        A string represetation of time according to NEXUS standard
    """
    # Convert to sting format recommented for NEXUS files
    t = ttime.strftime("%Y-%m-%dT%H:%M:%S+00:00", t)
    return t


def convert_time_from_nexus_string(t):
    """
    Convert time from NEXUS string to ``time.struct_time``

    Parameters
    ----------

    t : str
        A string represetation of time according to NEXUS standard

    Returns
    -------

    t : time.struct_time
        Time in the format returned by ``time.localtime`` or ``time.gmtime``
    """
    # Convert to sting format recommented for NEXUS files
    t = ttime.strptime(t, "%Y-%m-%dT%H:%M:%S+00:00")
    return t



from __future__ import absolute_import, division, print_function

import numpy as np
import json
from collections import OrderedDict
import copy
import math

from atom.api import Atom, Str, observe, Typed, Int, Dict, List, Float, Bool

from skbeam.fluorescence import XrfElement as Element
from skbeam.core.fitting.xrf_model import (
    ParamController,
    compute_escape_peak,
    trim,
    construct_linear_model,
    linear_spectrum_fitting,
)
from skbeam.core.fitting.xrf_model import K_LINE, L_LINE, M_LINE
from ..core.map_processing import snip_method_numba
from ..core.xrf_utils import check_if_eline_supported, get_eline_parameters, get_element_atomic_number

from ..core.utils import gaussian_sigma_to_fwhm, gaussian_fwhm_to_sigma

import logging

logger = logging.getLogger(__name__)


bound_options = ["none", "lohi", "fixed", "lo", "hi"]
fit_strategy_list = [
    "fit_with_tail",
    "free_more",
    "e_calibration",
    "linear",
    "adjust_element1",
    "adjust_element2",
    "adjust_element3",
]
autofit_param = ["e_offset", "e_linear", "fwhm_offset", "fwhm_fanoprime"]


class PreFitStatus(Atom):
    """
    Data structure for pre fit analysis.

    Attributes
    ----------
    z : str
        z number of element
    spectrum : array
        spectrum of given element
    status : bool
        True as plot is visible
    stat_copy : bool
        copy of status
    maxv : float
        max value of a spectrum
    norm : float
        norm value in respect to the strongest peak
    lbd_stat : bool
        define plotting status under a threshold value
    """

    z = Str()
    energy = Str()
    area = Float()
    spectrum = Typed(np.ndarray)
    status = Bool(False)
    stat_copy = Bool(False)
    maxv = Float()
    norm = Float()
    lbd_stat = Bool(False)


class ElementController(object):
    """
    This class performs basic ways to rank elements, show elements,
    calculate normed intensity, and etc.
    """

    def __init__(self):
        self.element_dict = OrderedDict()

    def delete_item(self, k):
        try:
            del self.element_dict[k]
            self.update_norm()
            logger.debug("Item {} is deleted.".format(k))
        except KeyError:
            pass

    def order(self, option="z"):
        """
        Order dict in different ways.
        """
        if option == "z":
            self.element_dict = OrderedDict(sorted(self.element_dict.items(), key=lambda t: t[1].z))
        elif option == "energy":
            self.element_dict = OrderedDict(sorted(self.element_dict.items(), key=lambda t: t[1].energy))
        elif option == "name":
            self.element_dict = OrderedDict(sorted(self.element_dict.items(), key=lambda t: t[0]))
        elif option == "maxv":
            self.element_dict = OrderedDict(
                sorted(self.element_dict.items(), key=lambda t: t[1].maxv, reverse=True)
            )

    def add_to_dict(self, dictv):
        """
        This function updates the dictionary element if it already exists.
        """
        self.element_dict.update(dictv)
        logger.debug("Item {} is added.".format(list(dictv.keys())))
        self.update_norm()

    def update_norm(self, threshv=0.0):
        """
        Calculate the normalized intensity for each element peak.

        Parameters
        ----------
        threshv : float
            No value is shown when smaller than the threshold value
        """
        # Do nothing if no elements are selected
        if not self.element_dict:
            return

        max_dict = np.max([v.maxv for v in self.element_dict.values()])

        for v in self.element_dict.values():
            v.norm = v.maxv / max_dict * 100
            v.lbd_stat = bool(v.norm > threshv)

        # also delete smaller values
        # there is some bugs in plotting when values < 0.0
        self.delete_peaks_below_threshold(threshv=threshv)

    def delete_all(self):
        self.element_dict.clear()

    def is_element_in_list(self, element_line_name):
        """
        Check if element 'k' is in the list of selected elements
        """
        if element_line_name in self.element_dict.keys():
            return True
        else:
            return False

    def get_element_list(self):
        current_elements = [v for v in self.element_dict.keys() if (v.lower() != v)]

        # logger.info('Current Elements for '
        #            'fitting are {}'.format(current_elements))
        return current_elements

    def update_peak_ratio(self):
        """
        If 'maxv' is modified, then the values of 'area' and 'spectrum' are adjusted accordingly:
        (1) maximum of spectrum is set equal to 'maxv'; (2) 'area' is scaled proportionally.
        It is important that only 'maxv' is changed before this function is called.
        """
        for v in self.element_dict.values():
            max_spectrum = np.max(v.spectrum)
            if not math.isclose(max_spectrum, 0.0, abs_tol=1e-20):
                factor = v.maxv / max_spectrum
            else:
                factor = 0.0
            v.spectrum *= factor
            v.area *= factor
        self.update_norm()

    def turn_on_all(self, option=True):
        """
        Set plotting status on for all lines.
        """
        if option is True:
            _plot = option
        else:
            _plot = False
        for v in self.element_dict.values():
            v.status = _plot

    def delete_peaks_below_threshold(self, threshv=0.1):
        """
        Delete elements smaller than threshold value. Non element
        peaks are not included.
        """
        remove_list = []
        non_element = ["compton", "elastic", "background"]
        for k, v in self.element_dict.items():
            if math.isnan(v.norm) or (v.norm >= threshv) or (k in non_element):
                continue
            # We don't want to delete userpeaks or pileup peaks (they are always added manually).
            if ("-" in k) or (k.lower().startswith("userpeak")):
                continue
            remove_list.append(k)
        for name in remove_list:
            del self.element_dict[name]
        return remove_list

    def delete_unselected_items(self):
        remove_list = []
        non_element = ["compton", "elastic", "background"]
        for k, v in self.element_dict.items():
            if k in non_element:
                continue
            if v.status is False:
                remove_list.append(k)
        for name in remove_list:
            del self.element_dict[name]
        return remove_list


class ParamModel(Atom):
    """
    The module used for maintain the set of fitting parameters.

    Attributes
    ----------
    parameters : `atom.Dict`
        A list of `Parameter` objects, subclassed from the `Atom` base class.
        These `Parameter` objects hold all relevant xrf information.
    data : array
        1D array of spectrum
    prefit_x : array
        xX axis with range defined by low and high limits.
    param_d : dict
        Parameters can be transferred into this dictionary.
    param_new : dict
        More information are saved, such as element position and width.
    total_y : dict
        Results from k lines
    total_y_l : dict
        Results from l lines
    total_y_m : dict
        Results from l lines
    e_list : str
        All elements used for fitting.
    file_path : str
        The path where file is saved.
    element_list : list
        The list of element lines selected for fitting
    n_selected_elines_for_fitting : Int
        The number of element lines selected for fitting
    n_selected_pure_elines_for_fitting : Int
        The number of element lines selected for fitting
            excluding pileup peaks and user defined peaks.
            Only 'pure' lines like Ca_K, K_K etc.
    """

    # Reference to FileIOModel object
    io_model = Typed(object)

    default_parameters = Dict()
    # data = Typed(np.ndarray)
    prefit_x = Typed(object)
    result_dict_names = List()
    param_new = Dict()
    total_y = Typed(object)
    # total_l = Dict()
    # total_m = Dict()
    # total_pileup = Dict()
    e_name = Str()  # Element line name selected in combo box
    add_element_intensity = Float(1000.0)
    element_list = List()
    # data_sets = Typed(OrderedDict)
    EC = Typed(object)
    x0 = Typed(np.ndarray)
    y0 = Typed(np.ndarray)
    max_area_dig = Int(2)
    auto_fit_all = Dict()
    bound_val = Float(1.0)

    energy_bound_high_buf = Float(0.0)
    energy_bound_low_buf = Float(0.0)

    n_selected_elines_for_fitting = Int(0)
    n_selected_pure_elines_for_fitting = Int(0)

    parameters_changed_cb = List()

    def __init__(self, *, default_parameters, io_model):
        try:
            self.io_model = io_model

            self.default_parameters = default_parameters
            self.param_new = copy.deepcopy(default_parameters)
            # TODO: do we set 'element_list' as a list of keys of 'EC.element_dict'
            self.element_list = get_element_list(self.param_new)
        except ValueError:
            logger.info("No default parameter files are chosen.")
        self.EC = ElementController()

        # The following line is part of the fix for automated updating of the energy bound
        #     in 'Automatic Element Finding' dialog box
        self.energy_bound_high_buf = self.param_new["non_fitting_values"]["energy_bound_high"]["value"]
        self.energy_bound_low_buf = self.param_new["non_fitting_values"]["energy_bound_low"]["value"]

    def add_parameters_changed_cb(self, cb):
        """
        Add callback to the list of callback function that are called after parameters are updated.
        """
        self.parameters_changed_cb.append(cb)

    def remove_parameters_changed_cb(self, cb):
        """
        Remove reference from the list of callback functions.
        """
        self.parameters_changed_cb = [_ for _ in self.parameters_changed_cb if _ != cb]

    def parameters_changed(self):
        """
        Run callback functions in the list. This method is expected to be called after the parameters
        are update to initiate necessary updates in the GUI.
        """
        for cb in self.parameters_changed_cb:
            cb()

    def default_param_update(self, default_parameters):
        """
        Replace the reference to the dictionary of default parameters.

        Parameters
        ----------
        default_parameters : dict
            Reference to complete and valid dictionary of default parameters.
        """
        self.default_parameters = default_parameters

    # The following function is part of the fix for automated updating of the energy bound
    #     in 'Automatic Element Finding' dialog box
    @observe("energy_bound_high_buf")
    def _update_energy_bound_high_buf(self, change):
        self.param_new["non_fitting_values"]["energy_bound_high"]["value"] = change["value"]
        self.define_range()

    @observe("energy_bound_low_buf")
    def _update_energy_bound_high_low(self, change):
        self.param_new["non_fitting_values"]["energy_bound_low"]["value"] = change["value"]
        self.define_range()

    def get_new_param_from_file(self, param_path):
        """
        Update parameters if new param_path is given.

        Parameters
        ----------
        param_path : str
            path to save the file
        """
        with open(param_path, "r") as json_data:
            self.param_new = json.load(json_data)
        self.create_spectrum_from_param_dict(reset=True)

        logger.info("Elements read from file are: {}".format(self.element_list))

    def update_new_param(self, param, reset=True):
        """
        Update the parameters based on the dictionary of parameters. Set ``reset=False``
        if selection status of elemental lines should be kept.

        Parameters
        ----------
        param : dict
            new dictionary of parameters
        reset : boolean
            reset (``True``) or clear (``False``) selection status of the element lines.
        """
        self.param_new = param
        self.create_spectrum_from_param_dict(reset=reset)

    @observe("bound_val")
    def _update_bound(self, change):
        if change["type"] != "create":
            logger.info(f"Peaks with values than the threshold {self.bound_val} will be removed from the list.")

    def define_range(self):
        """
        Cut x range according to values define in param_dict.
        """
        if self.io_model.data is None:
            return
        lowv = self.param_new["non_fitting_values"]["energy_bound_low"]["value"]
        highv = self.param_new["non_fitting_values"]["energy_bound_high"]["value"]
        self.x0, self.y0 = define_range(
            self.io_model.data,
            lowv,
            highv,
            self.param_new["e_offset"]["value"],
            self.param_new["e_linear"]["value"],
        )

    def create_spectrum_from_param_dict(self, reset=True):
        """
        Create spectrum profile with based on the current set of parameters.
        (``self.param_new`` -> ``self.EC`` and ``self.element_list``).
        Typical use: update self.param_new, then call this function.
        Set ``reset=False`` to keep selection status of the elemental lines.

        Parameters
        ----------
        reset : boolean
            clear or keep status of the elemental lines (in ``self.EC``).
        """
        param_dict = self.param_new
        self.element_list = get_element_list(param_dict)

        self.define_range()
        self.prefit_x, pre_dict, area_dict = calculate_profile(self.x0, self.y0, param_dict, self.element_list)
        # add escape peak
        if param_dict["non_fitting_values"]["escape_ratio"] > 0:
            pre_dict["escape"] = trim_escape_peak(self.io_model.data, param_dict, len(self.y0))

        temp_dict = OrderedDict()
        for e in pre_dict.keys():
            if e in ["background", "escape"]:
                spectrum = pre_dict[e]

                # summed spectrum here is not correct,
                # as the interval is assumed as 1, not energy interval
                # however area of background and escape is not used elsewhere, not important
                area = np.sum(spectrum)

                ps = PreFitStatus(
                    z=get_Z(e),
                    energy=get_energy(e),
                    area=float(area),
                    spectrum=spectrum,
                    maxv=float(np.around(np.max(spectrum), self.max_area_dig)),
                    norm=-1,
                    status=True,
                    lbd_stat=False,
                )
                temp_dict[e] = ps

            elif "-" in e:  # pileup peaks
                energy = self.get_pileup_peak_energy(e)
                energy = f"{energy:.4f}"
                spectrum = pre_dict[e]
                area = area_dict[e]

                ps = PreFitStatus(
                    z=get_Z(e),
                    energy=str(energy),
                    area=area,
                    spectrum=spectrum,
                    maxv=np.around(np.max(spectrum), self.max_area_dig),
                    norm=-1,
                    status=True,
                    lbd_stat=False,
                )
                temp_dict[e] = ps

            else:
                ename = e.split("_")[0]
                for k, v in param_dict.items():
                    energy = get_energy(e)  # For all peaks except Userpeaks

                    if ename in k and "area" in k:
                        spectrum = pre_dict[e]
                        area = area_dict[e]

                    elif ename == "compton" and k == "compton_amplitude":
                        spectrum = pre_dict[e]
                        area = area_dict[e]

                    elif ename == "elastic" and k == "coherent_sct_amplitude":
                        spectrum = pre_dict[e]
                        area = area_dict[e]

                    elif self.get_eline_name_category(ename) == "userpeak":
                        key = ename + "_delta_center"
                        energy = param_dict[key]["value"] + 5.0
                        energy = f"{energy:.4f}"
                    else:
                        continue

                    ps = PreFitStatus(
                        z=get_Z(ename),
                        energy=energy,
                        area=area,
                        spectrum=spectrum,
                        maxv=np.around(np.max(spectrum), self.max_area_dig),
                        norm=-1,
                        status=True,
                        lbd_stat=False,
                    )

                    temp_dict[e] = ps

        # Copy element status
        if not reset:
            element_status = {_: self.EC.element_dict[_].status for _ in self.EC.element_dict}

        self.EC.delete_all()
        self.EC.add_to_dict(temp_dict)

        if not reset:
            for key in self.EC.element_dict.keys():
                if key in element_status:
                    self.EC.element_dict[key].status = element_status[key]

        self.result_dict_names = list(self.EC.element_dict.keys())

    def get_selected_eline_energy_fwhm(self, eline):
        """
        Returns values of energy and fwhm for the peak 'eline' from the dictionary `self.param_new`.
        The emission line must exist in the dictionary. Primarily intended for use
        with user-defined peaks.

        Parameters
        ----------
        eline: str
            emission line (e.g. Ca_K) or peak name (e.g. Userpeak2, V_Ka1-Co_Ka1)
        """
        if eline not in self.EC.element_dict:
            raise ValueError(f"Emission line '{eline}' is not in the list of selected lines.")

        keys = self._generate_param_keys(eline)
        if not keys["key_dcenter"] or not keys["key_dsigma"]:
            raise ValueError(f"Failed to generate keys for the emission line '{eline}'.")

        energy = self.param_new[keys["key_dcenter"]]["value"] + 5.0
        dsigma = self.param_new[keys["key_dsigma"]]["value"]
        fwhm = gaussian_sigma_to_fwhm(dsigma) + self._compute_fwhm_base(energy)
        return energy, fwhm

    def get_pileup_peak_energy(self, eline):
        """
        Returns the energy (center) of pileup peak. Returns None if there is an error.

        Parameters
        ----------
        eline: str
            Name of the pileup peak, e.g. V_Ka1-Co_Ka1

        Returns
        -------
        float or None
            Energy in keV or None
        """
        incident_energy = self.param_new["coherent_sct_energy"]["value"]
        try:
            element_line1, element_line2 = eline.split("-")
            e1_cen = get_eline_parameters(element_line1, incident_energy)["energy"]
            e2_cen = get_eline_parameters(element_line2, incident_energy)["energy"]
            en = e1_cen + e2_cen
        except Exception:
            en = None
        return en

    def add_peak_manual(self, userpeak_center=2.5):
        """
        Manually add an emission line (or peak).

        Parameters
        ----------
        userpeak_center: float
            Center of the user defined peak. Ignored if emission line other
            than 'userpeak' is added
        """
        self._manual_input(userpeak_center=userpeak_center)
        self.update_name_list()
        self.data_for_plot()

    def remove_peak_manual(self):
        """
        Manually add an emission line (or peak). The name emission line (peak) to be deleted
        must be writtent to `self.e_name` before calling the function.
        """
        if self.e_name not in self.EC.element_dict:
            msg = (
                f"Line '{self.e_name}' is not in the list of selected lines,\n"
                f"therefore it can not be deleted from the list."
            )
            raise RuntimeError(msg)

        # Update parameter list
        self._remove_parameters_for_eline(self.e_name)

        # Update EC
        self.EC.delete_item(self.e_name)
        self.EC.update_peak_ratio()
        self.update_name_list()
        self.data_for_plot()

    def remove_elements_below_threshold(self, threshv=None):
        if threshv is None:
            threshv = self.bound_val

        deleted_elements = self.EC.delete_peaks_below_threshold(threshv=threshv)
        for eline in deleted_elements:
            self._remove_parameters_for_eline(eline)

        self.EC.update_peak_ratio()
        self.update_name_list()
        self.data_for_plot()

    def remove_elements_unselected(self):
        deleted_elements = self.EC.delete_unselected_items()
        for eline in deleted_elements:
            self._remove_parameters_for_eline(eline)

        self.EC.update_peak_ratio()
        self.update_name_list()
        self.data_for_plot()

    def _remove_parameters_for_eline(self, eline):
        """Remove entries for `eline` from the dictionary `self.param_new`"""
        if self.get_eline_name_category(eline) == "pileup":
            key_prefix = "pileup_" + self.e_name.replace("-", "_")
        else:
            key_prefix = eline

        # It is sufficient to compare using lowercase. It could be more reliable.
        key_prefix = key_prefix.lower()
        keys_to_delete = [_ for _ in self.param_new.keys() if _.lower().startswith(key_prefix)]
        for key in keys_to_delete:
            del self.param_new[key]

        # Add name to the name list
        _remove_element_from_list(eline, self.param_new)

    def _manual_input(self, userpeak_center=2.5):
        """
        Manually add an emission line (or peak).

        Parameters
        ----------
        userpeak_center: float
            Center of the user defined peak. Ignored if emission line other
            than 'userpeak' is added
        """

        if self.e_name in self.EC.element_dict:
            msg = f"Line '{self.e_name}' is in the list of selected lines. \nDuplicate entries are not allowed."
            raise RuntimeError(msg)

        default_area = 1e2

        # Add the new data entry to the parameter dictionary. This operation is necessary for 'userpeak'
        #   lines, because they need to be placed to the specific position (by setting 'delta_center'
        #   parameter, while regular element lines are placed to their default positions.
        d_energy = userpeak_center - 5.0

        # PC.params will contain a deepcopy of 'self.param_new' with the new line added
        PC = ParamController(self.param_new, [self.e_name])

        if self.get_eline_name_category(self.e_name) == "userpeak":
            energy = userpeak_center
            # Default values for 'delta_center'
            dc = copy.deepcopy(PC.params[f"{self.e_name}_delta_center"])
            # Modify the default values in the dictionary of parameters
            PC.params[f"{self.e_name}_delta_center"]["value"] = d_energy
            PC.params[f"{self.e_name}_delta_center"]["min"] = d_energy - (dc["value"] - dc["min"])
            PC.params[f"{self.e_name}_delta_center"]["max"] = d_energy + (dc["max"] - dc["value"])
        elif self.get_eline_name_category(self.e_name) == "pileup":
            energy = self.get_pileup_peak_energy(self.e_name)
        else:
            energy = get_energy(self.e_name)

        param_tmp = PC.params
        param_tmp = create_full_dict(param_tmp, fit_strategy_list)

        # Add name to the name list
        _add_element_to_list(self.e_name, param_tmp)

        # 'self.param_new' is used to provide 'hint' values for the model, but all active
        #    emission lines in 'elemental_lines' will be included in the model.
        #  The model will contain lines in 'elemental_lines', Compton and elastic
        x, data_out, area_dict = calculate_profile(
            self.x0, self.y0, param_tmp, elemental_lines=[self.e_name], default_area=default_area
        )

        # Check if element profile was calculated successfully.
        #   Calculation may fail if the selected line is not activated.
        #   The calculation is performed using ``xraylib` library, so there is no
        #   control over it.
        if self.e_name not in data_out:
            raise Exception(f"Failed to add the emission line '{self.e_name}': line is not activated.")

        # If model was generated successfully (the emission line was successfully added), then
        #   make temporary parameters permanent
        self.param_new = param_tmp

        ratio_v = self.add_element_intensity / np.max(data_out[self.e_name])

        ps = PreFitStatus(
            z=get_Z(self.e_name),
            energy=energy if isinstance(energy, str) else f"{energy:.4f}",
            area=area_dict[self.e_name] * ratio_v,
            spectrum=data_out[self.e_name] * ratio_v,
            maxv=self.add_element_intensity,
            norm=-1,
            status=True,  # for plotting
            lbd_stat=False,
        )

        self.EC.add_to_dict({self.e_name: ps})
        self.EC.update_peak_ratio()

    def _generate_param_keys(self, eline):
        """
        Returns prefix of the key from `param_new` dictionary based on emission line name
        If eline is actual emission line (like Ca_K), then the `key_dcenter` and `key_dsigma`
        point to 'a1' line (Ca_ka1). Function has to be extended if access to specific lines is
        required. Function is primarily intended for use with user-defined peaks.
        """

        category = self.get_eline_name_category(eline)
        if category == "pileup":
            eline = eline.replace("-", "_")
            key_area = "pileup_" + eline + "_area"
            key_dcenter = "pileup_" + eline + "delta_center"
            key_dsigma = "pileup_" + eline + "delta_sigma"
        elif category == "eline":
            eline = eline[:-1] + eline[-1].lower()
            key_area = eline + "a1_area"
            key_dcenter = eline + "a1_delta_center"
            key_dsigma = eline + "a1_delta_sigma"
        elif category == "userpeak":
            key_area = eline + "_area"
            key_dcenter = eline + "_delta_center"
            key_dsigma = eline + "_delta_sigma"
        elif eline == "compton":
            key_area = eline + "_amplitude"
            key_dcenter, key_dsigma = "", ""
        else:
            # No key exists (for "background", "escape", "elastic")
            key_area, key_dcenter, key_dsigma = "", "", ""
        return {"key_area": key_area, "key_dcenter": key_dcenter, "key_dsigma": key_dsigma}

    def modify_peak_height(self, maxv_new):
        """
        Modify the height of the emission line.

        Parameters
        ----------
        new_maxv: float
            New maximum value for the emission line `self.e_name`
        """

        ignored_peaks = {"escape"}
        if self.e_name in ignored_peaks:
            msg = f"Height of the '{self.e_name}' peak can not be changed."
            raise RuntimeError(msg)

        if self.e_name not in self.EC.element_dict:
            msg = (
                f"Attempt to modify maximum value for the emission line '{self.e_name},'\n"
                f"which is not currently selected."
            )
            raise RuntimeError(msg)

        key = self._generate_param_keys(self.e_name)["key_area"]
        maxv_current = self.EC.element_dict[self.e_name].maxv

        coef = maxv_new / maxv_current if maxv_current > 0 else 0
        # Only 'maxv' needs to be updated.
        self.EC.element_dict[self.e_name].maxv = maxv_new
        # The following function updates 'spectrum', 'area' and 'norm'.
        self.EC.update_peak_ratio()

        # Some of the parameters are represented only in EC, not in 'self.param_new'.
        #   (particularly "background" and "elastic")
        if key:
            self.param_new[key]["value"] *= coef

    def _compute_fwhm_base(self, energy):
        # Computes 'sigma' value based on default parameters and peak energy (for Userpeaks)
        #   does not include corrections for fwhm.
        # If both peak center (energy) and fwhm is updated, energy needs to be set first,
        #   since it is used in computation of ``fwhm_base``

        sigma = gaussian_fwhm_to_sigma(self.param_new["fwhm_offset"]["value"])

        sigma_sqr = energy + 5.0  # center
        sigma_sqr *= self.param_new["non_fitting_values"]["epsilon"]  # epsilon
        sigma_sqr *= self.param_new["fwhm_fanoprime"]["value"]  # fanoprime
        sigma_sqr += sigma * sigma  # We have computed the expression under sqrt

        sigma_total = np.sqrt(sigma_sqr)

        return gaussian_sigma_to_fwhm(sigma_total)

    def _update_userpeak_energy(self, eline, energy_new, fwhm_new):
        """
        Set new center for the user-defined peak at 'new_energy'
        """

        # According to the accepted peak model, as energy of the peak center grows,
        #   the peak becomes wider. The most user friendly solution is to automatically
        #   increase FWHM as the peak moves along the energy axis to the right and
        #   decrease otherwise. So generally, the user should first place the peak
        #   center at the desired energy, and then adjust FWHM.

        # We change energy, so we will have to change FWHM as well
        #  so before updating energy we will save the difference between
        #  the default (base) FWHM and the displayed FWHM

        name_userpeak_dcenter = eline + "_delta_center"
        old_energy = self.param_new[name_userpeak_dcenter]["value"]

        # This difference represents the required change in fwhm
        fwhm_difference = fwhm_new - self._compute_fwhm_base(old_energy)

        # Now we change energy.
        denergy = energy_new - 5.0

        v_center = self.param_new[name_userpeak_dcenter]["value"]
        v_max = self.param_new[name_userpeak_dcenter]["max"]
        v_min = self.param_new[name_userpeak_dcenter]["min"]
        # Keep the possible range for value change the same
        self.param_new[name_userpeak_dcenter]["value"] = denergy
        self.param_new[name_userpeak_dcenter]["max"] = denergy + v_max - v_center
        self.param_new[name_userpeak_dcenter]["min"] = denergy - (v_center - v_min)

        # The base value is updated now (since the energy has changed)
        fwhm_base = self._compute_fwhm_base(energy_new)
        fwhm = fwhm_difference + fwhm_base

        return fwhm

    def _update_userpeak_fwhm(self, eline, energy_new, fwhm_new):
        name_userpeak_dsigma = eline + "_delta_sigma"

        fwhm_base = self._compute_fwhm_base(energy_new)
        dfwhm = fwhm_new - fwhm_base

        dsigma = gaussian_fwhm_to_sigma(dfwhm)

        v_center = self.param_new[name_userpeak_dsigma]["value"]
        v_max = self.param_new[name_userpeak_dsigma]["max"]
        v_min = self.param_new[name_userpeak_dsigma]["min"]
        # Keep the possible range for value change the same
        self.param_new[name_userpeak_dsigma]["value"] = dsigma
        self.param_new[name_userpeak_dsigma]["max"] = dsigma + v_max - v_center
        self.param_new[name_userpeak_dsigma]["min"] = dsigma - (v_center - v_min)

    def _update_userpeak_energy_fwhm(self, eline, fwhm_new, energy_new):
        """
        Update energy and fwhm of the user-defined peak 'eline'. The 'delta_center'
        and 'delta_sigma' parameters in the `self.param_new` dictionary are updated.
        `area` should be updated after call to this function. This function also
        doesn't change entries in the `EC` dictionary.
        """
        # Ensure, that the values are greater than some small value to ensure that
        #   there is no computational problems.
        # Energy resolution for the existing beamlines is 0.01 keV, so 0.001 is small
        #   enough both for center energy and FWHM.
        energy_small_value = 0.001
        energy_new = max(energy_new, energy_small_value)
        fwhm_new = max(fwhm_new, energy_small_value)

        fwhm_new = self._update_userpeak_energy(eline, energy_new, fwhm_new)
        self._update_userpeak_fwhm(eline, energy_new, fwhm_new)

    def modify_userpeak_params(self, maxv_new, fwhm_new, energy_new):

        if self.get_eline_name_category(self.e_name) != "userpeak":
            msg = (
                f"Hight and width can be modified only for a user defined peak.\n"
                f"The function was called for '{self.e_name}' peak"
            )
            raise RuntimeError(msg)

        if self.e_name not in self.EC.element_dict:
            msg = (
                f"Attempt to modify maximum value for the emission line '{self.e_name},'\n"
                f"which is not currently selected."
            )
            raise RuntimeError(msg)

        # Some checks of the input values
        if maxv_new <= 0.0:
            raise ValueError("Peak height must be a positive number greater than 0.001.")
        if energy_new <= 0.0:
            raise ValueError("User peak energy must be a positive number greater than 0.001.")
        if fwhm_new <= 0:
            raise ValueError("User peak FWHM must be a positive number.")

        # Make sure that the energy of the user peak is within the selected fitting range
        energy_bound_high = self.param_new["non_fitting_values"]["energy_bound_high"]["value"]
        energy_bound_low = self.param_new["non_fitting_values"]["energy_bound_low"]["value"]
        if energy_new > energy_bound_high or energy_new < energy_bound_low:
            raise ValueError("User peak energy is outside the selected range.")

        # This updates 'delta_center' and 'delta_sigma' entries of the 'self.param_new' dictionary
        self._update_userpeak_energy_fwhm(self.e_name, fwhm_new, energy_new)

        default_area = 1e2
        key = self._generate_param_keys(self.e_name)["key_area"]

        # Set area to default area, change it later once the area is computed
        self.param_new[key]["value"] = default_area

        # 'self.param_new' is used to provide 'hint' values for the model, but all active
        #    emission lines in 'elemental_lines' will be included in the model.
        #  The model will contain lines in 'elemental_lines', Compton and elastic
        x, data_out, area_dict = calculate_profile(
            self.x0, self.y0, self.param_new, elemental_lines=[self.e_name], default_area=default_area
        )

        ratio_v = maxv_new / np.max(data_out[self.e_name])

        area = area_dict[self.e_name] * ratio_v
        self.param_new[key]["value"] = area

        ps = PreFitStatus(
            z=get_Z(self.e_name),
            energy=f"{energy_new:.4f}",
            area=area,
            spectrum=data_out[self.e_name] * ratio_v,
            maxv=maxv_new,
            norm=-1,
            status=True,  # for plotting
            lbd_stat=False,
        )

        self.EC.element_dict[self.e_name] = ps

        logger.debug(
            f"The parameters of the user defined peak. The new values:\n"
            f"   Energy: {energy_new} keV, FWHM: {fwhm_new}, Maximum: {maxv_new}\n"
        )

    def generate_pileup_peak_name(self, name1, name2):
        """
        Returns name for the pileup peak. The element line with the lowest
        energy is placed first in the name.
        """
        incident_energy = self.param_new["coherent_sct_energy"]["value"]
        e1 = get_eline_parameters(name1, incident_energy)["energy"]
        e2 = get_eline_parameters(name2, incident_energy)["energy"]
        if e1 > e2:
            name1, name2 = name2, name1
        return name1 + "-" + name2

    def update_name_list(self):
        """
        When result_dict_names change, the looper in enaml will update.
        """
        # need to clean list first, in order to refresh the list in GUI
        self.result_dict_names = []
        self.result_dict_names = list(self.EC.element_dict.keys())
        self.element_list = get_element_list(self.param_new)

        peak_list = self.get_user_peak_list()
        # Create the list of selected emission lines such as Ca_K, K_K, etc.
        #   No pileup or user peaks
        pure_peak_list = [n for n in self.result_dict_names if n in peak_list]
        self.n_selected_elines_for_fitting = len(self.result_dict_names)
        self.n_selected_pure_elines_for_fitting = len(pure_peak_list)

        logger.info(f"The update list of emission lines: {self.result_dict_names}")

    def get_eline_name_category(self, eline_name):
        """
        Returns the category to which `eline_name` belongs: `eline`, `userpeak`,
        `pileup` or `other`.

        Parameters
        ----------
        eline_name: str
            Name to be analyzed

        Returns
        -------
        str
            category: one of `("eline", "userpeak", "pileup" or "other")`
        """
        if check_if_eline_supported(eline_name):
            return "eline"
        elif eline_name.lower().startswith("userpeak"):
            return "userpeak"
        elif "-" in eline_name:  # This is specific to currently accepted naming convention
            return "pileup"
        else:
            return "other"

    def _sort_eline_list(self, element_list):
        """
        Sort the list of elements
        """
        names_elines, names_userpeaks, names_pileup_peaks, names_other = [], [], [], []
        for name in element_list:
            if self.get_eline_name_category(name) == "eline":
                try:
                    z = get_element_atomic_number(name.split("_")[0])
                except Exception:
                    z = 0
                names_elines.append([name, z])
            elif self.get_eline_name_category(name) == "userpeak":
                names_userpeaks.append(name)
            elif self.get_eline_name_category(name) == "pileup":
                names_pileup_peaks.append(name)
            else:
                names_other.append(name)

        names_elines.sort(key=lambda v: int(v[1]))  # Sort by Z (atomic number)
        names_elines = [_[0] for _ in names_elines]  # Get rid of Z
        names_userpeaks.sort()
        names_pileup_peaks.sort()
        names_other.sort()

        return names_elines + names_userpeaks + names_pileup_peaks + names_other

    def get_sorted_result_dict_names(self):
        """
        The function returns the list of selected emission lines. The emission lines are
        sorted in the following order: emission line names (sorted in the order of growing
        atomic number Z), userpeaks (in alphabetic order), pileup peaks (in alphabetic order),
        other peaks (in alphabetic order).

        Returns
        -------
        list(str)
            the list if emission line names
        """
        return self._sort_eline_list(self.result_dict_names)

    def get_sorted_element_list(self):
        """
        Returns sorted ``element_list``.
        """
        return self._sort_eline_list(self.element_list)

    def read_param_from_file(self, param_path):
        """
        Update parameters if new param_path is given.

        Parameters
        ----------
        param_path : str
            path to save the file
        """
        with open(param_path, "r") as json_data:
            param = json.load(json_data)
            self.update_new_param(param, reset=True)

    def find_peak(self, *, threshv=0.1, elemental_lines=None):
        """
        Run automatic peak finding, and save results as dict of object.

        Parameters
        ----------
        threshv: float
            The value will not be shown on GUI if it is smaller than the threshold.

        elemental_lines: list(str)
            The list of elemental lines to find. If ``None``, then all supported
            lines (K, L and M) are searched.
        """
        self.define_range()  # in case the energy calibraiton changes
        self.prefit_x, out_dict, area_dict = linear_spectrum_fitting(
            self.x0, self.y0, self.param_new, elemental_lines=elemental_lines
        )
        logger.info(
            "Energy range: {}, {}".format(
                self.param_new["non_fitting_values"]["energy_bound_low"]["value"],
                self.param_new["non_fitting_values"]["energy_bound_high"]["value"],
            )
        )

        prefit_dict = OrderedDict()
        for k, v in out_dict.items():
            ps = PreFitStatus(
                z=get_Z(k),
                energy=get_energy(k),
                area=area_dict[k],
                spectrum=v,
                maxv=np.around(np.max(v), self.max_area_dig),
                norm=-1,
                lbd_stat=False,
            )
            prefit_dict.update({k: ps})

        logger.info("Automatic Peak Finding found elements as : {}".format(list(prefit_dict.keys())))
        self.EC.delete_all()
        self.EC.add_to_dict(prefit_dict)

        self.create_full_param()

    def create_full_param(self):
        """
        Update current ``self.param_new`` with elements from ``self.EC`` (delete elements that
        are not in ``self.EC`` and update the existing elements.
        """
        self.define_range()
        # We set 'self.element_list' from 'EC' (because we want to set elements of 'self.param_new'
        #   from 'EC.element_dict'
        self.element_list = self.EC.get_element_list()

        self.param_new = update_param_from_element(self.param_new, self.element_list)
        element_temp = [e for e in self.element_list if len(e) <= 4]
        pileup_temp = [e for e in self.element_list if "-" in e]
        userpeak_temp = [e for e in self.element_list if "user" in e.lower()]

        # update area values in param_new according to results saved in ElementController
        if len(self.EC.element_dict):
            for k, v in self.param_new.items():
                if "area" in k:
                    if "pileup" in k:
                        name_cut = k[7:-5]  # remove pileup_ and _area
                        for p in pileup_temp:
                            if name_cut == p.replace("-", "_"):
                                v["value"] = self.EC.element_dict[p].area
                    elif "user" in k.lower():
                        for p in userpeak_temp:
                            if p in k:
                                v["value"] = self.EC.element_dict[p].area
                    else:
                        for e in element_temp:
                            k_name, k_line, _ = k.split("_")
                            e_name, e_line = e.split("_")
                            if k_name == e_name and e_line.lower() == k_line[0]:  # attention: S_k and As_k
                                v["value"] = self.EC.element_dict[e].area

            if "compton" in self.EC.element_dict:
                self.param_new["compton_amplitude"]["value"] = self.EC.element_dict["compton"].area
            if "coherent_sct_amplitude" in self.EC.element_dict:
                self.param_new["coherent_sct_amplitude"]["value"] = self.EC.element_dict["elastic"].area

            if "escape" in self.EC.element_dict:
                self.param_new["non_fitting_values"]["escape_ratio"] = self.EC.element_dict[
                    "escape"
                ].maxv / np.max(self.y0)
            else:
                self.param_new["non_fitting_values"]["escape_ratio"] = 0.0

    def data_for_plot(self):
        """
        Save data in terms of K, L, M lines for plot.
        """
        self.total_y = None
        self.auto_fit_all = {}

        for k, v in self.EC.element_dict.items():
            if v.status is True:
                self.auto_fit_all[k] = v.spectrum
                if self.total_y is None:
                    self.total_y = np.array(v.spectrum)  # need to copy an array
                else:
                    self.total_y += v.spectrum

        # for k, v in new_dict.items():
        #     if '-' in k:  # pileup
        #         self.total_pileup[k] = self.EC.element_dict[k].spectrum
        #     elif 'K' in k:
        #         self.total_y[k] = self.EC.element_dict[k].spectrum
        #     elif 'L' in k:
        #         self.total_l[k] = self.EC.element_dict[k].spectrum
        #     elif 'M' in k:
        #         self.total_m[k] = self.EC.element_dict[k].spectrum
        #     else:
        #         self.total_y[k] = self.EC.element_dict[k].spectrum

    def get_user_peak_list(self):
        """
        Returns the list of element emission peaks
        """
        return K_LINE + L_LINE + M_LINE

    def get_selected_emission_line_data(self):
        """
        Assembles the full emission line data for processing.

        Returns
        -------
        list(dict)
            Each dictionary includes the following data: "name" (e.g. Ca_ka1 etc.),
            "area" (estimated peak area based on current fitting results), "ratio"
            (ratio such as Ca_ka2/Ca_ka1)
        """
        # Full list of supported emission lines (such as Ca_K)
        supported_elines = self.get_user_peak_list()
        # Parameter keys start with full emission line name (eg. Ca_ka1)
        param_keys = list(self.param_new.keys())

        incident_energy = self.param_new["coherent_sct_energy"]["value"]

        full_line_list = []
        for eline in self.EC.element_dict.keys():
            if eline not in supported_elines:
                continue
            area = self.EC.element_dict[eline].area
            lines = [_ for _ in param_keys if _.lower().startswith(eline.lower())]
            lines = set(["_".join(_.split("_")[:2]) for _ in lines])
            for ln in lines:
                eline_info = get_eline_parameters(ln, incident_energy)
                data = {"name": ln, "area": area, "ratio": eline_info["ratio"], "energy": eline_info["energy"]}
                full_line_list.append(data)
        return full_line_list

    def guess_pileup_peak_components(self, energy, tolerance=0.05):
        """
        Provides a guess on components of pileup peak based on the set of selected emission lines,
        and selected energy.

        Parameters
        ----------
        energy: float
            Approximate (selected) energy of pileup peak location
        tolerance: float
            Allowed deviation of the sum of component energies from the selected energy, keV

        Returns
        -------
        tuple(str, str, float)
            Component emission lines (such as Ca_ka1, K_ka1 etc) and the energy of
            the resulting pileup peak.
        """

        line_data = self.get_selected_emission_line_data()
        energy_min, energy_max = energy - tolerance, energy + tolerance

        # Not very efficient algorithm, which tries all combinations of lines
        pileup_components, areas = [], []
        for n1, line1 in enumerate(line_data):
            for n2 in range(n1, len(line_data)):
                line2 = line_data[n2]
                if energy_min < line1["energy"] + line2["energy"] < energy_max:
                    if line1 == line2:
                        area = line1["area"] * line1["ratio"]
                    else:
                        area = line1["area"] * line1["ratio"] + line2["area"] * line2["ratio"]
                    pileup_components.append((line1["name"], line2["name"], line1["energy"] + line2["energy"]))
                    areas.append(area)

        if len(areas):
            # Find index with maximum area
            n = areas.index(max(areas))
            return pileup_components[n]
        else:
            return None


def save_as(file_path, data):
    """
    Save full param dict into a file.
    """
    with open(file_path, "w") as outfile:
        json.dump(data, outfile, sort_keys=True, indent=4)


def define_range(data, low, high, a0, a1):
    """
    Cut x range according to values define in param_dict.

    Parameters
    ----------
    data : array
        raw spectrum
    low : float
        low bound in KeV
    high : float
        high bound in KeV
    a0 : float
        offset term of energy calibration
    a1 : float
        linear term of energy calibration

    Returns
    -------
    x : array
        trimmed channel number
    y : array
        trimmed spectrum according to x
    """
    x = np.arange(data.size)

    #  ratio to transfer energy value back to channel value
    # approx_ratio = 100

    low_new = int(np.around((low - a0) / a1))
    high_new = int(np.around((high - a0) / a1))
    x0, y0 = trim(x, data, low_new, high_new)
    return x0, y0


def calculate_profile(x, y, param, elemental_lines, default_area=1e5):
    """
    Calculate the spectrum profile based on given parameters. Use function
    construct_linear_model from xrf_model.

    Parameters
    ----------
    x : array
        channel array
    y : array
        spectrum intensity
    param : dict
        parameters
    elemental_lines : list
        such as Si_K, Pt_M
    default_area : float
        default value for the gaussian area of each element

    Returns
    -------
    x : array
        trimmed energy range
    temp_d : dict
        dict of array
    area_dict : dict
        dict of area for elements and other peaks
    """
    # Need to use deepcopy here to avoid unexpected change on parameter dict
    fitting_parameters = copy.deepcopy(param)

    total_list, matv, area_dict = construct_linear_model(
        x, fitting_parameters, elemental_lines, default_area=default_area
    )

    temp_d = {k: v for (k, v) in zip(total_list, matv.transpose())}

    # add background
    bg = snip_method_numba(
        y,
        fitting_parameters["e_offset"]["value"],
        fitting_parameters["e_linear"]["value"],
        fitting_parameters["e_quadratic"]["value"],
        width=fitting_parameters["non_fitting_values"]["background_width"],
    )
    temp_d["background"] = bg

    x_energy = (
        fitting_parameters["e_offset"]["value"]
        + fitting_parameters["e_linear"]["value"] * x
        + fitting_parameters["e_quadratic"]["value"] * x ** 2
    )

    return x_energy, temp_d, area_dict


def trim_escape_peak(data, param_dict, y_size):
    """
    Calculate escape peak within required range.

    Parameters
    ----------
    data : array
        raw spectrum
    param_dict : dict
        parameters for fitting
    y_size : int
        the size of trimmed spectrum

    Returns
    -------
    array :
        trimmed escape peak spectrum
    """
    ratio = param_dict["non_fitting_values"]["escape_ratio"]
    xe, ye = compute_escape_peak(data, ratio, param_dict)
    lowv = param_dict["non_fitting_values"]["energy_bound_low"]["value"]
    highv = param_dict["non_fitting_values"]["energy_bound_high"]["value"]
    xe, es_peak = trim(xe, ye, lowv, highv)
    logger.info("Escape peak is considered with ratio {}".format(ratio))

    # align to the same length
    if y_size > es_peak.size:
        temp = es_peak
        es_peak = np.zeros(y_size)
        es_peak[: temp.size] = temp
    else:
        es_peak = es_peak[:y_size]
    return es_peak


def create_full_dict(param, name_list, fixed_list=["adjust_element2", "adjust_element3"]):
    """
    Create full param dict so each item has the same nested dict.
    This is for GUI purpose only.

    Pamameters
    ----------
    param : dict
        all parameters including element
    name_list : list
        strategy names

    Returns
    -------
    dict: with update
    """
    param_new = copy.deepcopy(param)
    for n in name_list:
        for k, v in param_new.items():
            if k == "non_fitting_values":
                continue
            if n not in v:
                # enforce newly created parameter to be fixed
                # for strategy in fixed_list
                if n in fixed_list:
                    v.update({n: "fixed"})
                else:
                    v.update({n: v["bound_type"]})
    return param_new


def strip_line(ename):
    return ename.split("_")[0]


def get_Z(ename):
    """
    Return element's Z number.

    Parameters
    ----------
    ename : str
        element name

    Returns
    -------
    int or None
        element Z number
    """

    non_element = ["compton", "elastic", "background", "escape"]
    if (ename.lower() in non_element) or "-" in ename or "user" in ename.lower():
        return "-"
    else:
        e = Element(strip_line(ename))
        return str(e.Z)


def get_energy(ename):
    """
    Return energy value by given elemental name. Need to consider non-elemental case.
    """
    non_element = ["compton", "elastic", "background", "escape"]
    if (ename.lower() in non_element) or "user" in ename.lower():
        return "-"
    else:
        e = Element(strip_line(ename))
        ename = ename.lower()
        if "_k" in ename:
            energy = e.emission_line["ka1"]
        elif "_l" in ename:
            energy = e.emission_line["la1"]
        elif "_m" in ename:
            energy = e.emission_line["ma1"]

        return str(np.around(energy, 4))


def get_element_list(param):
    """Extract elements from parameter class object"""
    element_list = param["non_fitting_values"]["element_list"]
    element_list = [e.strip(" ") for e in element_list.split(",")]
    # Unfortunately, "".split(",") returns [""] instead of [], but we need [] !!!
    if element_list == [""]:
        element_list = []
    return element_list


def _set_element_list(element_list, param):
    element_list = ", ".join(element_list)
    param["non_fitting_values"]["element_list"] = element_list


def _add_element_to_list(eline, param):
    """Add element to list in the parameter class object"""
    elist = get_element_list(param)
    elist_lower = [_.lower() for _ in elist]
    if eline.lower() not in elist_lower:
        elist.append(eline)
    _set_element_list(elist, param)


def _remove_element_from_list(eline, param):
    """Add element to list in the parameter class object"""
    elist = get_element_list(param)
    elist_lower = [_.lower() for _ in elist]
    try:
        index = elist_lower.index(eline.lower())
        elist.pop(index)
        _set_element_list(elist, param)
    except ValueError:
        pass


def param_dict_cleaner(parameter, element_list):
    """
    Make sure param only contains element from element_list.

    Parameters
    ----------
    parameter : dict
        fitting parameters
    element_list : list
        list of elemental lines

    Returns
    -------
    dict :
        new param dict containing given elements
    """
    param = copy.deepcopy(parameter)
    param_new = {}

    elines_list = [e for e in element_list if len(e) <= 4]
    elines_lower = [e.lower() for e in elines_list]
    pileup_list = [e for e in element_list if "-" in e]
    userpeak_list = [e for e in element_list if "user" in e.lower()]

    new_element_set = set()

    for k, v in param.items():
        if k == "non_fitting_values" or k == k.lower():
            param_new.update({k: v})
        elif "pileup" in k:
            for p in pileup_list:
                if p.replace("-", "_") in k:
                    param_new.update({k: v})
                    new_element_set.add(p)
        elif "user" in k.lower():
            for p in userpeak_list:
                if p in k:
                    param_new.update({k: v})
                    new_element_set.add(p)
        elif k[:3].lower() in elines_lower:
            index = elines_lower.index(k[:3].lower())
            param_new.update({k: v})
            new_element_set.add(elines_list[index])
        elif k[:4].lower() in elines_lower:
            index = elines_lower.index(k[:4].lower())
            param_new.update({k: v})
            new_element_set.add(elines_list[index])

    new_element_list = list(new_element_set)
    _set_element_list(new_element_list, param_new)

    return param_new


def update_param_from_element(param, element_list):
    """
    Clean up or extend param according to new element list.

    Parameters
    ----------
    param : dict
        fitting parameters
    element_list : list
        list of elemental lines

    Returns
    -------
    dict
    """
    param_new = copy.deepcopy(param)

    for eline in element_list:
        _add_element_to_list(eline, param_new)

    # first remove some items not included in element_list
    param_new = param_dict_cleaner(param_new, element_list)

    # second add some elements to a full parameter dict
    # create full parameter list including elements
    PC = ParamController(param_new, element_list)
    # parameter values not updated based on param_new, so redo it
    param_temp = PC.params

    #  enforce adjust_element area to be fixed,
    #  while bound_type in xrf_model is defined as none for area
    # for k, v in param_temp.items():
    #     if '_area' in k:
    #         v['bound_type'] = 'fixed'

    for k, v in param_temp.items():
        if k == "non_fitting_values":
            continue
        if k in param_new:
            param_temp[k] = param_new[k]
            # for k1 in v.keys():
            #     v[k1] = param_new[k][k1]
    param_new = param_temp

    # to create full param dict, for GUI only
    param_new = create_full_dict(param_new, fit_strategy_list)
    return param_new



from qtpy.QtWidgets import QHBoxLayout, QVBoxLayout, QLabel, QDialog, QDialogButtonBox, QRadioButton, QButtonGroup
from qtpy.QtGui import QIntValidator, QRegExpValidator
from qtpy.QtCore import QRegExp

from .useful_widgets import LineEditExtended, set_tooltip, IntValidatorStrict

import logging

logger = logging.getLogger(__name__)


class DialogSelectScan(QDialog):
    def __init__(self):

        super().__init__()

        self.resize(400, 200)
        self.setWindowTitle("Load Run From Database")

        self._id_uid = None
        self._mode_id_uid = "id"

        label = QLabel("Enter run ID or UID:")
        self.le_id_uid = LineEditExtended()
        self.le_id_uid.textChanged.connect(self.le_id_uid_text_changed)
        self.le_id_uid.editingFinished.connect(self.le_id_uid_editing_finished)
        set_tooltip(self.le_id_uid, "Enter <b>Run ID</b> or <b>Run UID</b>.")

        self._validator_id = IntValidatorStrict()
        # Short UID example: "04c9afa7"
        self._validator_uid_short = QRegExpValidator(QRegExp(r"[0-9a-f]{8}"))
        # Full UID example: "04c9afa7-a43a-4af1-8e55-2034384d4a77"
        self._validator_uid_full = QRegExpValidator(
            QRegExp(r"[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}")
        )

        self.rb_id = QRadioButton("Run ID")
        set_tooltip(self.rb_id, "The value in the line edit box is <b>Run ID</b> (e.g. <b>34235</b> or <b>-1</b>)")
        self.rb_id.setChecked(self._mode_id_uid == "id")
        self.rb_uid = QRadioButton("Run UID")
        self.rb_uid.setChecked(self._mode_id_uid == "uid")
        set_tooltip(
            self.rb_uid,
            "The value in the line edit box is <b>Run UID</b> "
            "(e.g. <b>04c9afb7-a43a-4af1-8e55-2034384d4a77</b> or <b>04c9afb7</b>)",
        )

        self.btn_group = QButtonGroup()
        self.btn_group.addButton(self.rb_id)
        self.btn_group.addButton(self.rb_uid)
        self.btn_group.buttonToggled.connect(self.btn_group_button_toggled)

        button_box = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        button_box.accepted.connect(self.accept)
        button_box.rejected.connect(self.reject)
        self.pb_ok = button_box.button(QDialogButtonBox.Ok)

        vbox = QVBoxLayout()

        vbox.addStretch(1)

        hbox = QHBoxLayout()
        hbox.addWidget(label)
        hbox.addStretch(1)
        vbox.addLayout(hbox)
        vbox.addWidget(self.le_id_uid)

        hbox = QHBoxLayout()
        hbox.addStretch(1)
        hbox.addWidget(self.rb_id)
        hbox.addStretch(1)
        hbox.addWidget(self.rb_uid)
        hbox.addStretch(1)
        vbox.addLayout(hbox)

        vbox.addStretch(1)
        vbox.addWidget(button_box)
        self.setLayout(vbox)

        # This is how the button from QDialogButtonBox can be disabled
        # button_box.button(QDialogButtonBox.Ok).setEnabled(False)

        self.le_id_uid.setText("")
        self._validate_id_uid()

    def btn_group_button_toggled(self, button, state):
        if state:
            if button == self.rb_id:
                self._mode_id_uid = "id"
            elif button == self.rb_uid:
                self._mode_id_uid = "uid"

            text = self.le_id_uid.text()
            if self._validate_id_uid(text):
                self._read_id_uid(text)

    def le_id_uid_text_changed(self, text):
        self._validate_id_uid(text)

    def le_id_uid_editing_finished(self):
        text = self.le_id_uid.text()
        if self._validate_id_uid(text):
            self._read_id_uid(text)

    def get_id_uid(self):
        """
        Read the selected scan ID or UID

        Returns
        -------
        (str, int) or (str, str)
            keyword "id" or "uid" depending on whether the second element is scan ID or UID,
            scan ID (int), scan UID (str) or None if no scan ID or UID is selected
        """
        return self._mode_id_uid, self._id_uid

    def _validate_id_uid(self, text=None):
        valid = False
        if text is None:
            text = self.le_id_uid.text()
        if self._mode_id_uid == "id":
            if self._validator_id.validate(text, 0)[0] == QIntValidator.Acceptable:
                valid = True
        elif self._mode_id_uid == "uid":
            if (
                self._validator_uid_short.validate(text, 0)[0] == QIntValidator.Acceptable
                or self._validator_uid_full.validate(text, 0)[0] == QIntValidator.Acceptable
            ):
                valid = True

        self.le_id_uid.setValid(valid)
        self.pb_ok.setEnabled(valid)

        return valid

    def _read_id_uid(self, text):
        # It is assumed that the entered text is valid for the selected mode
        if text is None:
            text = self.le_id_uid.text()
        if self._mode_id_uid == "id":
            self._id_uid = int(text)
        elif self._mode_id_uid == "uid":
            self._id_uid = text

    def _set_mode(self, mode):
        if mode == "id":
            self._mode_id_uid = mode
            self.rb_id.setChecked(True)
        elif mode == "uid":
            self._mode_id_uid = mode
            self.rb_uid.setChecked(True)



from qtpy.QtWidgets import (
    QWidget,
    QVBoxLayout,
    QHBoxLayout,
    QRadioButton,
    QButtonGroup,
    QComboBox,
    QCheckBox,
    QPushButton,
    QDialog,
)
from qtpy.QtCore import Qt, Slot, Signal

from matplotlib.backends.backend_qt5agg import (
    FigureCanvasQTAgg as FigureCanvas,
    NavigationToolbar2QT as NavigationToolbar,
)

from .useful_widgets import ElementSelection, set_tooltip, global_gui_variables

from .dlg_plot_escape_peak import DialogPlotEscapePeak


class PlotFittingModel(QWidget):

    signal_selected_element_changed = Signal(str)
    signal_add_line = Signal()
    signal_remove_line = Signal()

    def __init__(self, *, gpc, gui_vars):
        super().__init__()

        self._enable_events = False

        # Global processing classes
        self.gpc = gpc
        # Global GUI variables (used for control of GUI state)
        self.gui_vars = gui_vars

        self.combo_plot_type = QComboBox()
        self.combo_plot_type.addItems(["LinLog", "Linear"])
        # Values are received and sent to the model, the don't represent the displayed text
        self.combo_plot_type_values = ["linlog", "linear"]

        self.cb_show_spectrum = QCheckBox("Show spectrum")

        self.cb_show_fit = QCheckBox("Show fit")

        self.rb_selected_region = QRadioButton("Selected region")
        self.rb_selected_region.setChecked(True)

        self.rb_full_spectrum = QRadioButton("Full spectrum")

        self.bgroup = QButtonGroup()
        self.bgroup.addButton(self.rb_selected_region)
        self.bgroup.addButton(self.rb_full_spectrum)

        self.pb_escape_peak = QPushButton("Escape Peak ...")
        self.pb_escape_peak.clicked.connect(self.pb_escape_peak_clicked)

        self.pb_add_line = QPushButton("Add Line")
        self.pb_add_line.clicked.connect(self.pb_add_line_clicked)

        self.pb_remove_line = QPushButton("Remove Line")
        self.pb_remove_line.clicked.connect(self.pb_remove_line_clicked)

        self.element_selection = ElementSelection()
        eline_sample_list = ["Li_K", "B_K", "C_K", "N_K", "Fe_K", "Userpeak1"]
        self.element_selection.set_item_list(eline_sample_list)
        self.element_selection.signal_current_item_changed.connect(self.element_selection_item_changed)

        self.mpl_canvas = FigureCanvas(self.gpc.plot_model._fig)
        self.mpl_toolbar = NavigationToolbar(self.mpl_canvas, self)

        # Keep layout without change when canvas is hidden (invisible)
        sp_retain = self.mpl_canvas.sizePolicy()
        sp_retain.setRetainSizeWhenHidden(True)
        self.mpl_canvas.setSizePolicy(sp_retain)

        self.widgets_enable_events(True)

        vbox = QVBoxLayout()

        hbox = QHBoxLayout()
        hbox.addWidget(self.combo_plot_type)
        hbox.addWidget(self.cb_show_spectrum)
        hbox.addWidget(self.cb_show_fit)
        hbox.addStretch(1)
        hbox.addWidget(self.rb_selected_region)
        hbox.addWidget(self.rb_full_spectrum)
        vbox.addLayout(hbox)

        hbox = QHBoxLayout()
        hbox.addWidget(self.pb_escape_peak)
        hbox.addStretch(1)
        hbox.addWidget(self.pb_add_line)
        hbox.addWidget(self.pb_remove_line)
        hbox.addSpacing(20)
        hbox.addWidget(self.element_selection)
        hbox.addStretch(1)
        vbox.addLayout(hbox)

        vbox.addWidget(self.mpl_toolbar)
        vbox.addWidget(self.mpl_canvas)
        self.setLayout(vbox)

        self._set_tooltips()

    def widgets_enable_events(self, enable):
        if enable:
            if not self._enable_events:
                self.cb_show_spectrum.toggled.connect(self.cb_show_spectrum_toggled)
                self.cb_show_fit.toggled.connect(self.cb_show_fit_toggled)
                self.bgroup.buttonToggled.connect(self.bgroup_button_toggled)
                self.combo_plot_type.currentIndexChanged.connect(self.combo_plot_type_current_index_changed)
                self._enable_events = True
        else:
            if self._enable_events:
                self.cb_show_spectrum.toggled.disconnect(self.cb_show_spectrum_toggled)
                self.cb_show_fit.toggled.disconnect(self.cb_show_fit_toggled)
                self.bgroup.buttonToggled.disconnect(self.bgroup_button_toggled)
                self.combo_plot_type.currentIndexChanged.disconnect(self.combo_plot_type_current_index_changed)
                self._enable_events = False

    def _set_tooltips(self):
        set_tooltip(self.combo_plot_type, "Use <b>Linear</b> or <b>LinLog</b> axes to plot spectra")
        set_tooltip(self.cb_show_spectrum, "Show <b>raw<b> spectrum")
        set_tooltip(self.cb_show_fit, "Show <b>fitted</b> spectrum")
        set_tooltip(
            self.rb_selected_region,
            "Plot spectrum in the <b>selected range</b> of energies. The range may be set "
            "in the 'Model' tab. Click the button <b>'Find Automatically ...'</b> "
            "to set the range of energies before finding the emission lines. The range "
            "may be changed in General Settings dialog (button <b>'General ...'</b>) at any time.",
        )
        set_tooltip(self.rb_full_spectrum, "Plot full spectrum over <b>all available eneriges</b>.")
        set_tooltip(
            self.pb_escape_peak,
            "Select options for displaying the <b>escape peak</b>. "
            "If activated, the location of the escape peak is shown "
            "for the emission line, which is currently selected for adding, "
            "removal or editing.",
        )
        set_tooltip(self.pb_add_line, "<b>Add</b> the current emission line to the list of selected lines")
        set_tooltip(
            self.pb_remove_line, "<b>Remove</b> the current emission line from the list of selected lines."
        )
        set_tooltip(self.element_selection, "<b>Choose</b> the emission line for addition or removal.")

    def update_widget_state(self, condition=None):
        if condition == "tooltips":
            self._set_tooltips()
        self.mpl_toolbar.setVisible(self.gui_vars["show_matplotlib_toolbar"])

        # Hide Matplotlib canvas during computations
        state_compute = global_gui_variables["gui_state"]["running_computations"]
        self.mpl_canvas.setVisible(not state_compute)

    @Slot()
    def update_controls(self):
        self.widgets_enable_events(False)
        plot_spectrum, plot_fit = self.gpc.get_line_plot_state()
        self.cb_show_spectrum.setChecked(plot_spectrum)
        self.cb_show_fit.setChecked(plot_fit)
        if self.gpc.get_plot_fit_energy_range() == "selected":
            self.rb_selected_region.setChecked(True)
        else:
            self.rb_full_spectrum.setChecked(True)

        try:
            index = self.combo_plot_type_values.index(self.gpc.get_plot_fit_linlog())
            self.combo_plot_type.setCurrentIndex(index)
        except ValueError:
            self.combo_plot_type.setCurrentIndex(-1)

        self.widgets_enable_events(True)

    def le_mouse_press(self, event):
        print("Button pressed (line edit)")
        if event.button() == Qt.RightButton:
            print("Right button pressed")

    def pb_escape_peak_clicked(self, event):
        plot_escape_peak, detector_material = self.gpc.get_escape_peak_params()
        incident_energy = self.gpc.get_incident_energy()
        dlg = DialogPlotEscapePeak()
        dlg.set_parameters(plot_escape_peak, incident_energy, detector_material)
        if dlg.exec() == QDialog.Accepted:
            plot_escape_peak, detector_material = dlg.get_parameters()
            self.gpc.set_escape_peak_params(plot_escape_peak, detector_material)
            print("Dialog exit: Ok button")

    def cb_show_spectrum_toggled(self, state):
        self.gpc.show_plot_spectrum(state)

    def cb_show_fit_toggled(self, state):
        self.gpc.show_plot_fit(state)

    def bgroup_button_toggled(self, button, checked):
        if checked:
            if button == self.rb_selected_region:
                self.gpc.set_plot_fit_energy_range("selected")
            else:
                self.gpc.set_plot_fit_energy_range("full")

    def combo_plot_type_current_index_changed(self, index):
        self.gpc.set_plot_fit_linlog(self.combo_plot_type_values[index])

    def pb_add_line_clicked(self):
        self.signal_add_line.emit()

    def pb_remove_line_clicked(self):
        self.signal_remove_line.emit()

    def element_selection_item_changed(self, index, eline):
        self.signal_selected_element_changed.emit(eline)

    @Slot(str)
    def slot_selection_item_changed(self, eline):
        self.element_selection.set_current_item(eline)

    @Slot()
    def slot_update_eline_selection_list(self):
        eline_list = self.gpc.get_full_eline_list()
        self.element_selection.set_item_list(eline_list)

    @Slot(bool, bool)
    def slot_update_add_remove_btn_state(self, add_enabled, remove_enabled):
        self.pb_add_line.setEnabled(add_enabled)
        self.pb_remove_line.setEnabled(remove_enabled)

    @Slot()
    @Slot(bool)
    def redraw_plot_fit(self):
        self.gpc.update_plot_fit()



import numpy as np
from scipy.optimize import nnls


def rfactor_compute(spectrum, fit_results, ref_spectra):
    r"""
    Computes R-factor for the fitting results

    Parameters
    ----------
    spectrum : ndarray
        spectrum data on which fitting is performed (N elements)

    fit_results : ndarray
        results of fitting (coefficients, K elements)

    ref_spectra : 2D ndarray
        reference spectra used for fitting (NxK element array)

    Returns
    -------
        float, the value of R-factor
    """

    # Check if input parameters are valid
    assert (
        spectrum.ndim == 1 or spectrum.ndim == 2
    ), "Parameter 'spectrum' must be 1D or 2D array, ({spectrum.ndim})"
    assert spectrum.ndim == fit_results.ndim, (
        f"Spectrum data (ndim = {spectrum.ndim}) and fitting results "
        f"(ndim = {fit_results.ndim}) must have the same number of dimensions"
    )
    assert ref_spectra.ndim == 2, "Parameter 'ref_spectra' must be 2D array, ({ref_spectra.ndim})"
    assert spectrum.shape[0] == ref_spectra.shape[0], (
        f"Arrays 'spectrum' ({spectrum.shape}) and 'ref_spectra' ({ref_spectra.shape}) "
        "must have the same number of data points"
    )
    assert fit_results.shape[0] == ref_spectra.shape[1], (
        f"Arrays 'fit_results' ({fit_results.shape}) and 'ref_spectra' ({ref_spectra.shape}) "
        "must have the same number of spectrum points"
    )
    if spectrum.ndim == 2:  # Only if multiple spectra are processed
        assert spectrum.shape[1] == fit_results.shape[1], (
            f"Arrays 'spectrum' {spectrum.shape} and 'fit_results' {fit_results.shape}"
            "must have the same number of columns"
        )

    spectrum_fit = np.matmul(ref_spectra, fit_results)
    return rfactor(spectrum, spectrum_fit)


def rfactor(spectrum_experimental, spectrum_fit):
    r"""
    Computes R-factor based on two spectra

    Parameters
    ----------
    spectrum_experimental : ndarray
        spectrum data on which fitting is performed (N elements)

    spectrum_fit : ndarray
        fitted spectrum (weighted sum of spectrum components, N elements)

    Returns
    -------
        float, the value of R-factor
    """

    # Compute R-factor
    dif = spectrum_experimental - spectrum_fit
    dif_sum = np.sum(np.abs(dif), axis=0)
    data_sum = np.sum(np.abs(spectrum_experimental), axis=0)

    # Avoid accidental division by zero (or a very small number)
    data_sum = np.clip(data_sum, a_min=1e-30, a_max=None)

    return dif_sum / data_sum


def fit_spectrum(data, ref_spectra, *, method="nnls", axis=0, maxiter=100, rate=0.2, epsilon=1e-30):
    r"""
    Perform fitting of a single or or multiple spectra. A single spectra is represented as
    a 1D ndarray. Multiple spectra collected from scan of a line, 2D or 3D image may be represented
    as multidimensional array containing spectral information along axis ``axis``. The returned
    fitting result have the same dimensionality as the input data with fitted coefficients
    located along the same axis as the spectrum data.

    Parameters
    ----------
    data : ndarray
        single or multidimensional spectral data. If fitting done for a single spectrum with K energy points,
        ``data`` is a 1D ndarray with K points. If 'data' contains spectra obtained by scanning NxM map,
        ``data`` is 3D ndarray with dimensions (K,N,M), (N,K,M) or (N,M,K). The axis containing spectra
        is specified by the parameter ``axis``, which should be set to 0 (default), 1 or 2 for the for the
        3D example above. Dimensionality of ``data`` array is not restricted and the function will perform
        properly if ``axis`` points to the correct dimension.

    ref_spectra : ndarray (2D)
        array with columns representing the reference spectra. If fitting is performed for Q reference
        spectra, then ``ref_spectra`` has dimensions (K,Q), where K is the number of energy points.

    method : str
        optimization method used for fitting. Currently supported methods are "nnls" and "admm".

    axis : int
        the number of the axis in the ``data`` array that hold the spectral information. If ``data``
        array has ``n`` dimensions, then ``axis`` may take values in the range ``-n .. n-1``. The
        fitting results will be placed in the output data array along the same axis.

    maxiter : int
        maximum number of iterations. Optimization may stop prematurely if convergence criteria are met.

    rate : float
        descent rate for optimization algorithm. Currently is used only for ADMM fitting (1/lambda).

    epsilon : float
        small value used in stopping criterion of ADMM optimization algorithm.

    Returns
    -------
    weights : ndarray
        array with the same number of dimensions as ``data``. Weights are placed along the axis ``axis``.
        For example, if ``data`` has the shape (N,K,M) and ``axis=1``, then ``weights`` has
        the shape (N,Q,M), where Q is the number of spectrum references.

    rfactor : ndarray
        array that contains R-factor value for each fitted spectrum. For example, if ``data`` has shape
        (N,K,M) and ``axis=1``, then ``rfactor`` has shape (N,M)

    results_dict : dict
        dictionary that contains additional information returned by a fitting routine. The contents of
        the dictionary depends on the optimization routine. Common information: ``method`` - string
        that contains the name of the optimization method (``nnls`` or ``admm``).

        NNLS optimization:

            - ``residual`` - an array that contains values of least-squares difference between the observed
            and fitted spectra. The same shape as ``rfactor``.

        ADMM optimization:

            - ``convergence`` - 1D array of values that represent change of weights at each iteration of
            the optimization. The number of elements is equal to the number of iterations.

            - ``feasibility`` - 1D array, same shape as ``convergence``
    """
    # Explicitly check if one of the supported optimisation method is specified
    method = method.lower()
    supported_fitting_methods = ("nnls", "admm")
    assert (
        method in supported_fitting_methods
    ), f"Fitting method '{method}' is not supported. Supported methods: {supported_fitting_methods}"

    data = np.asarray(data)
    ref_spectra = np.asarray(ref_spectra)

    assert ref_spectra.ndim == 2, f"The array 'ref_spectra' must have 2 dimensions instead of {ref_spectra.ndim}"

    assert (axis >= -data.ndim) and (
        axis < data.ndim
    ), f"Specified axis {axis} does not exist in data array. Allowed values: {-data.ndim} .. {data.ndim - 1}"

    # Switch to standard data view (spectrum points along axis==0)
    data = np.moveaxis(data, axis, 0)

    n_pts = data.shape[0]
    data_dims = data.shape[1:]
    n_pts_2 = ref_spectra.shape[0]
    n_refs = ref_spectra.shape[1]

    assert (
        n_pts == n_pts_2
    ), f"The number of spectrum points in data ({n_pts}) and references ({n_pts_2}) do not match."

    assert rate > 0.0, f"The parameter 'rate' is zero or negative ({rate:.6g})"

    assert maxiter > 0, f"The parameter 'maxiter' is zero or negative ({rate})"

    assert epsilon > 0.0, f"The parameter 'epsilon' is zero or negative ({rate:.6g})"

    # Depending on 'data_dim', there could be three cases
    #   'data_dim' is empty - data is 1D array representing a single point, 1D array of weights
    #                         will be returned, data must be converted to 2D array for processing
    #   'data_dim' has one element - data is 2D array, representing one line of pixels, process as is
    #   'data_dim' has more than one element - data is multidimensional array representing
    #                        multidimensional image, reshape to 1D data (2D array) for processing
    #                        and the convert back to multidimensional representation

    if not data_dims:
        data_1D = np.expand_dims(data, axis=1)
    elif len(data_dims) > 1:
        data_1D = np.reshape(data, [n_pts, np.prod(data_dims)])
    else:
        data_1D = data

    if method == "admm":
        weights, rfactor, convergence, feasibility = _fitting_admm(
            data_1D, ref_spectra, rate=rate, maxiter=maxiter, epsilon=epsilon
        )
    else:
        # Call the default "nnls" method, since this is the only choice left.
        weights, rfactor, residual = _fitting_nnls(data_1D, ref_spectra, maxiter=maxiter)

    # Reshape the fitting results in case of a single pixel or multidimensional map
    if not data_dims:
        weights = np.squeeze(weights, axis=1)
        rfactor = rfactor[0]
        if method == "nnls":
            residual = residual[0]
    elif len(data_dims) > 1:
        weights = np.reshape(weights, np.insert(data_dims, 0, n_refs))
        rfactor = np.reshape(rfactor, data_dims)
        if method == "nnls":
            residual = np.reshape(residual, data_dims)

    # Now swap back the results of fitting (coefficients are along the same axis as the spectrum points)
    weights = np.moveaxis(weights, 0, axis)

    # The results returned for each optimization method include 'weights' and 'rfactor'
    #   Additional results are different for each optimisation method and are returned as
    #   elements of the dictionary ``results_dict``.
    if method == "admm":
        results_dict = {
            "method": method,
            "convergence": convergence,
            "feasibility": feasibility,
        }
    else:
        results_dict = {
            "method": method,
            "residual": residual,
        }

    return weights, rfactor, results_dict


def _fitting_nnls(data, ref_spectra, *, maxiter=100):
    r"""
    Fitting of multiple spectra using NNLS method.

    Parameters
    ----------

    data : ndarray(float), 2D
        array holding multiple observed spectra, shape (K, N), where K is the number of energy points,
        and N is the number of spectra

    absorption_refs : ndarray(float), 2D
        array of references, shape (K, Q), where Q is the number of references.

    maxiter : int
        maximum number of iterations. Optimization may stop prematurely if convergence criteria are met.

    Returns
    -------

    map_data_fitted : ndarray(float), 2D
        fitting results, shape (Q, N), where Q is the number of references and N is the number of spectra.

    map_rfactor : ndarray(float), 2D
        map that represents R-factor for the fitting, shape (M,N).

    map_residual : ndarray(float), 2D
        residual returned by NNLS algorithm
    """
    assert data.ndim == 2, "Data array 'data' must have 2 dimensions"
    assert ref_spectra.ndim == 2, "Data array 'ref_spectra' must have 2 dimensions"

    n_pts = data.shape[0]
    n_pixels = data.shape[1]
    n_pts_2 = ref_spectra.shape[0]
    n_refs = ref_spectra.shape[1]

    assert (
        n_pts == n_pts_2
    ), f"The number of spectrum points in data ({n_pts}) and references ({n_pts_2}) do not match."

    assert maxiter > 0, f"The parameter 'maxiter' is zero or negative ({maxiter})"

    map_data_fitted = np.zeros(shape=[n_refs, n_pixels])
    map_rfactor = np.zeros(shape=[n_pixels])
    map_residual = np.zeros(shape=[n_pixels])
    for n in range(n_pixels):
        map_sel = data[:, n]
        result, residual = nnls(ref_spectra, map_sel, maxiter=maxiter)

        rfactor = rfactor_compute(map_sel, result, ref_spectra)

        map_data_fitted[:, n] = result
        map_rfactor[n] = rfactor
        map_residual[n] = residual

    return map_data_fitted, map_rfactor, map_residual


def _fitting_admm(data, ref_spectra, *, rate=0.2, maxiter=100, epsilon=1e-30, non_negative=True):
    r"""
    Fitting of multiple spectra using ADMM method.

    Parameters
    ----------

    data : ndarray(float), 2D
        array holding multiple observed spectra, shape (K, N), where K is the number of energy points,
        and N is the number of spectra

    absorption_refs : ndarray(float), 2D
        array of references, shape (K, Q), where Q is the number of references.

    maxiter : int
        maximum number of iterations. Optimization may stop prematurely if convergence criteria are met.

    rate : float
        descent rate for optimization algorithm. Currently is used only for ADMM fitting (1/lambda).

    epsilon : float
        small value used in stopping criterion of ADMM optimization algorithm.

    non_negative : bool
        if True, then the solution is guaranteed to be non-negative

    Returns
    -------

    map_data_fitted : ndarray(float), 2D
        fitting results, shape (Q, N), where Q is the number of references and N is the number of spectra.

    map_rfactor : ndarray(float), 2D
        map that represents R-factor for the fitting, shape (M,N).

    convergence : ndarray(float), 1D
        convergence data returned by ADMM algorithm

    feasibility : ndarray(float), 1D
        feasibility data returned by ADMM algorithm

    The prototype for the ADMM fitting function was implemented by Hanfei Yan in Matlab.
    """
    assert data.ndim == 2, "Data array 'data' must have 2 dimensions"
    assert ref_spectra.ndim == 2, "Data array 'ref_spectra' must have 2 dimensions"

    n_pts = data.shape[0]
    n_pixels = data.shape[1]
    n_pts_2 = ref_spectra.shape[0]
    n_refs = ref_spectra.shape[1]

    assert (
        n_pts == n_pts_2
    ), f"ADMM fitting: number of spectrum points in data ({n_pts}) and references ({n_pts_2}) do not match."

    assert rate > 0.0, f"ADMM fitting: parameter 'rate' is zero or negative ({rate:.6g})"

    assert maxiter > 0, f"ADMM fitting: parameter 'maxiter' is zero or negative ({rate})"

    assert epsilon > 0.0, f"ADMM fitting: parameter 'epsilon' is zero or negative ({rate:.6g})"

    y = data
    # Calculate some quantity to be used in the iteration
    A = ref_spectra
    At = np.transpose(A)

    z = np.matmul(At, y)
    c = np.matmul(At, A)

    # Initialize variables
    w = np.ones(shape=[n_refs, n_pixels])
    u = np.zeros(shape=[n_refs, n_pixels])

    # Feasibility test: x == w
    convergence = np.zeros(shape=[maxiter])
    feasibility = np.zeros(shape=[maxiter])

    dg = np.eye(n_refs, dtype=float) * rate
    m1 = np.linalg.inv((c + dg))

    n_iter = 0
    for i in range(maxiter):
        m2 = z + (w - u) * rate
        x = np.matmul(m1, m2)
        w_updated = x + u
        if non_negative:
            w_updated = w_updated.clip(min=0)
        u = u + x - w_updated

        conv = np.linalg.norm(w_updated - w) / np.linalg.norm(w_updated)
        convergence[i] = conv
        feasibility[i] = np.linalg.norm(x - w_updated)

        w = w_updated

        if conv < epsilon:
            n_iter = i + 1
            break

    # Compute R-factor
    rfactor = rfactor_compute(data, w, ref_spectra)

    convergence = convergence[:n_iter]
    feasibility = feasibility[:n_iter]

    return w, rfactor, convergence, feasibility
